# file: jax2onnx/converter/function_handling.py


from __future__ import annotations

from collections.abc import Callable
import inspect
import logging
import re
from typing import TYPE_CHECKING
import numpy as np  # Add missing numpy import

import jax.numpy as jnp
from jax.core import ShapedArray
from jax.extend.core import Literal
from onnx import helper
from onnx import TensorProto  # noqa: F401 – used by helpers
import onnx

from jax2onnx.converter.name_generator import get_qualified_name
from jax2onnx.converter.onnx_builder import OnnxBuilder

logger = logging.getLogger("jax2onnx.converter.function_handling")

if TYPE_CHECKING:  # circular‑import guard
    from jax2onnx.converter.jaxpr_converter import Jaxpr2OnnxConverter


def create_scalar_constant_tensor(param_name, param_value, dtype_enum, parent_builder):
    # Use the parameter name directly for better clarity and to pass the test
    const_name = param_name

    # Check if we already have this constant
    for initializer in parent_builder.initializers:
        if initializer.name == const_name:
            logger.debug(
                f"Using existing constant tensor '{const_name}' for parameter '{param_name}'"
            )
            return const_name

    const_tensor = onnx.helper.make_tensor(
        name=const_name,
        data_type=dtype_enum,
        dims=(),
        vals=[int(param_value) if isinstance(param_value, bool) else param_value],
    )
    parent_builder.initializers.append(const_tensor)
    logger.debug(
        f"Created constant tensor '{const_name}' for parameter '{param_name}' with value {param_value}"
    )
    return const_name


def prepare_function_names(converter, orig_fn, name):
    impl_key = get_qualified_name(orig_fn)
    logger.debug(f"Encountered function primitive: {impl_key}")

    unique_node_name = converter.builder.get_unique_instance_name(name.split(".")[-1])
    logger.debug(f"Generating unique ONNX node name: {unique_node_name}")

    parent_builder = converter.builder
    return impl_key, unique_node_name, parent_builder


def check_parameters(
    name: str, converter: "Jaxpr2OnnxConverter", eqn, orig_fn: Callable, params
):
    if orig_fn is None:
        raise RuntimeError(f"Original function for {name} not recorded.")


def resolve_function_inputs(converter, eqn, parent_builder):
    input_names, example_args, outer_input_vars_avals = [], [], []
    for var in eqn.invars:
        if hasattr(var, "aval"):
            aval, var_name = var.aval, converter.get_name(var)
            input_names.append(var_name)
            outer_input_vars_avals.append((var, aval))
            example_args.append(create_example_arg(aval))
            register_input_metadata(parent_builder, var_name, aval)
        elif isinstance(var, Literal):
            example_args.append(var.val)
        else:
            raise TypeError(f"Unexpected input var type: {type(var)}")
    return input_names, example_args, outer_input_vars_avals


def create_example_arg(aval):
    """Create an example argument for function tracing.

    Handles both concrete shapes and shapes with symbolic dimensions.
    """
    if not aval.shape:
        return jnp.zeros((), dtype=aval.dtype)

    # Check if the shape contains any symbolic dimensions
    has_symbolic_dim = any(not isinstance(dim, int) for dim in aval.shape)

    if has_symbolic_dim:
        # Create a placeholder ShapedArray instead of concrete array
        # This avoids trying to materialize arrays with symbolic dimensions
        return aval
    else:
        # For concrete shapes, create actual array
        return jnp.ones(aval.shape, dtype=aval.dtype)


def register_input_metadata(builder, var_name, aval):
    shape, dtype = tuple(aval.shape), aval.dtype
    builder.register_value_info_metadata(var_name, shape, dtype)


def register_constant_parameter(
    const_name, param_name, param_value, input_names, extra_param_inputs, example_args
):
    input_names.append(const_name)
    extra_param_inputs.append((param_name, const_name))
    example_args.append(param_value)


def process_scalar_parameters(
    scalar_params_to_process,
    converter,
    eqn,
    parent_builder,
    input_names,
    extra_param_inputs,
    example_args,
    special_param_handling=None,  # New argument for explicit parameter handling
):
    """
    Process scalar parameters for function conversion.
    special_param_handling: Optional[dict], maps param_name to one of: 'input', 'constant', 'static'.
    """
    special_param_handling = special_param_handling or {}
    for param_name, param_value in scalar_params_to_process.items():
        if any(name == param_name for name, _ in extra_param_inputs):
            continue

        handling_mode = special_param_handling.get(param_name, None)

        # If the parameter is a tracer (not static), expose as ONNX input and skip all constant logic
        is_tracer = str(type(param_value)).find("DynamicJaxprTracer") >= 0
        if is_tracer:
            logger.debug(
                f"Exposing tracer parameter '{param_name}' as ONNX input (no constant created)."
            )
            if param_name not in input_names:
                input_names.append(param_name)
            extra_param_inputs.append((param_name, param_name))
            # Do NOT create a constant or add to example_args
            continue

        if handling_mode == "input":
            # Treat as graph input
            if param_name in converter.name_to_var:
                var_name = param_name
                logger.debug(
                    f"Using existing graph input '{var_name}' for parameter '{param_name}' (explicit input mode)"
                )
                if var_name not in input_names:
                    input_names.append(var_name)
                extra_param_inputs.append((param_name, var_name))
                continue
        elif handling_mode == "constant":
            # Treat as constant
            # use INT32 for plain Python ints to match JAX's default
            dtype_enum = (
                onnx.TensorProto.BOOL
                if isinstance(param_value, bool)
                else (
                    onnx.TensorProto.INT32
                    if isinstance(param_value, int)
                    else onnx.TensorProto.FLOAT
                )
            )
            const_name = create_scalar_constant_tensor(
                param_name, param_value, dtype_enum, parent_builder
            )
            register_constant_parameter(
                const_name,
                param_name,
                param_value,
                input_names,
                extra_param_inputs,
                example_args,
            )
            continue
        elif handling_mode == "static":
            # Do not add as input or constant, just skip
            logger.debug(
                f"Parameter '{param_name}' marked as static, skipping input/constant registration."
            )
            continue

        # Fallback: legacy behavior for backward compatibility
        # Remove all hardcoded param_name checks
        if isinstance(param_value, (bool, int, float)):
            dtype_enum = (
                onnx.TensorProto.BOOL
                if isinstance(param_value, bool)
                else (
                    onnx.TensorProto.INT32
                    if isinstance(param_value, int)
                    else onnx.TensorProto.FLOAT
                )
            )
            const_name = create_scalar_constant_tensor(
                param_name, param_value, dtype_enum, parent_builder
            )
            register_constant_parameter(
                const_name,
                param_name,
                param_value,
                input_names,
                extra_param_inputs,
                example_args,
            )
        else:
            input_names.append(param_name)
            extra_param_inputs.append((param_name, param_value))
            logger.warning(
                f"Unsupported parameter type for {param_name}: {type(param_value)}"
            )
            example_args.append(param_value)


def classify_scalar_params(
    params, static_params, eqn, converter, input_names, extra_param_inputs, orig_fn=None
):
    """
    Classify scalar parameters for ONNX function export.

    Required (no default) parameters are always exposed as ONNX inputs.
    Optional parameters (with a default) are exported as ONNX constants if possible.
    Uses Python's inspect to determine which parameters are required.
    Tracer values are treated as dynamic inputs unless a static value is allowed.
    """
    scalar_params_to_process = {}
    # Determine which parameters are required (no default) using inspect
    required_params = set()
    if orig_fn is not None:
        sig = inspect.signature(orig_fn)
        for name, param in sig.parameters.items():
            if param.default is inspect.Parameter.empty and param.kind in (
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.KEYWORD_ONLY,
            ):
                required_params.add(name)
    for param_name, param_value in params.items():
        is_tracer = str(type(param_value)).find("DynamicJaxprTracer") >= 0
        # Only use static value if parameter is NOT required (i.e., has a default)
        if is_tracer:
            if param_name in static_params and (
                orig_fn is None or param_name not in required_params
            ):
                # Optional parameter: use static value
                param_value = static_params[param_name]
            # else: required parameter, expose as ONNX input
        if isinstance(param_value, (bool, int, float)) or is_tracer:
            scalar_params_to_process[param_name] = param_value

            # Remove legacy logic for deterministic/training guessing
            # Always expose as input if it's a tracer and not static
            # (No more bool_input_indices guessing)
    return scalar_params_to_process


def handle_function_parameters(
    params, converter, eqn, parent_builder, input_names, example_args, orig_fn=None
):
    extra_param_inputs = []
    if not params:
        return extra_param_inputs

    static_params = getattr(converter, "static_params", {})
    scalar_params_to_process = classify_scalar_params(
        params,
        static_params,
        eqn,
        converter,
        input_names,
        extra_param_inputs,
        orig_fn=orig_fn,
    )
    process_scalar_parameters(
        scalar_params_to_process,
        converter,
        eqn,
        parent_builder,
        input_names,
        extra_param_inputs,
        example_args,
    )
    return extra_param_inputs


def prepare_trace_kwargs_and_example_args(params, example_args):
    trace_kwargs = {"preserve_graph": True}
    param_keys_to_exclude = []

    if params:
        trace_kwargs["params"] = params
        param_keys_to_exclude = list(params.keys())
        logger.debug(
            f"Will exclude these parameters from example_args: {param_keys_to_exclude}"
        )

        if example_args:
            if (
                isinstance(example_args[-1], bool)
                and "deterministic" in param_keys_to_exclude
            ):
                logger.debug(
                    "Removing duplicated 'deterministic' parameter from example_args"
                )
                example_args = example_args[:-1]

            for i, param_name in enumerate(param_keys_to_exclude):
                if param_name in [
                    "mask",
                    "dropout_rng",
                    "dtype",
                    "precision",
                    "module",
                ] and i < len(example_args):
                    if example_args[i] is None:
                        logger.debug(
                            f"Removing duplicated '{param_name}' parameter from example_args"
                        )
                        example_args = example_args[:i] + example_args[i + 1 :]

    return trace_kwargs, example_args


def propagate_eqn_parameters(eqn, params):
    if eqn.params:
        if params is None:
            params = {}
        for param_key, param_value in eqn.params.items():
            if param_key not in params:
                params[param_key] = param_value
                logger.debug(
                    f"Propagating parameter '{param_key}' from equation params"
                )
    return params


def setup_sub_converter(converter, eqn, params, unique_node_name, parent_builder):
    """Set up a sub-converter for handling function bodies.

    This function creates a child ONNX builder and converter, ensuring symbolic
    dimensions are properly propagated from parent to child.
    """
    logger.debug(
        "⇢ enter @onnx_function  (parent symbols: %s)",
        (
            parent_builder.var_to_symbol_map
            if hasattr(parent_builder, "var_to_symbol_map")
            else {}
        ),
    )

    # ------------------------------------------------------------------ #
    # 1.  Create a *child* builder, but seed it with the symbol aliases   #
    #     already discovered for the outer graph so that the original     #
    #     symbolic names ('B', etc.) survive inside the Function.         #
    # ------------------------------------------------------------------ #
    sub_builder = OnnxBuilder(
        parent_builder.name_generator,
        parent_builder.opset,
        unique_node_name + "_graph",
        initializers=parent_builder.initializers,
        converter=converter,  # Pass converter reference
    )

    # ─────────────────────────  DEBUG / bookkeeping  ────────────────
    if not hasattr(sub_builder, "var_to_symbol_map"):
        sub_builder.var_to_symbol_map = {}

    if hasattr(parent_builder, "var_to_symbol_map"):
        sub_builder.var_to_symbol_map.update(parent_builder.var_to_symbol_map)

    for k, v in list(sub_builder.var_to_symbol_map.items()):
        sub_builder.var_to_symbol_map.setdefault(str(k), v)
        sub_builder.var_to_symbol_map.setdefault(v, v)
    logger.debug("   inherited symbols → %s", sub_builder.var_to_symbol_map)
    # ────────────────────────────────────────────────────────────────

    # Propagate other symbolic dimension related maps
    if hasattr(parent_builder, "symbolic_shapes") and hasattr(
        sub_builder, "symbolic_shapes"
    ):
        sub_builder.symbolic_shapes.update(parent_builder.symbolic_shapes)

    # Optional: keep a link so the Function can fall back to parent map for new aliases
    sub_builder.parent = parent_builder

    # ------------------------------------------------------------------ #
    # 2.  Create the converter for the Function and propagate symbolic    #
    #     dimension mapping information                                   #
    # ------------------------------------------------------------------ #

    # ──► Create the *Function*-local JaxprConverter and hand over
    #     its private symbol tables
    sub_converter = converter.__class__(sub_builder)

    # share Var → 'B' table
    if hasattr(converter, "_dimvar_to_name"):
        # Start with a copy of parent's dimension mapping
        sub_converter._dimvar_to_name = dict(converter._dimvar_to_name)

    # share the canonical tuple of abstracted axes
    if hasattr(converter, "symbolic_axes"):
        sub_converter.symbolic_axes = converter.symbolic_axes

    # make sure to enable shape polymorphism in the sub-converter
    if hasattr(converter, "use_abstracted_axes"):
        sub_converter.use_abstracted_axes = converter.use_abstracted_axes

    # (optional) allow the child converter to fall back to the parent
    sub_converter.parent = converter

    # Propagate equation parameters
    params = propagate_eqn_parameters(eqn, params)
    sub_converter.params = params

    # Propagate other parameters if available
    if hasattr(converter, "call_params"):
        sub_converter.call_params = converter.call_params

    return sub_converter, sub_builder, params


def register_function_inputs(
    sub_converter, sub_builder, internal_input_vars, outer_input_vars_avals
):
    for internal_var, (outer_var, outer_aval) in zip(
        internal_input_vars[: len(outer_input_vars_avals)],
        outer_input_vars_avals,
        strict=False,
    ):
        internal_name = sub_converter.get_name(internal_var)
        shape = tuple(outer_aval.shape)
        dtype = helper.np_dtype_to_tensor_dtype(outer_aval.dtype)
        sub_builder.register_value_info_metadata(
            internal_name, shape, dtype, origin="function_input"
        )
        sub_builder.add_value_info(internal_name, shape, dtype)


def rename_and_register_param_inputs(
    sub_converter, sub_builder, remaining_internal_vars, extra_param_inputs
):
    """
    Ensure that parameter inputs in the ONNX function graph use descriptive names
    (matching the original function parameter names) instead of generic variable names.
    This improves readability and correctness, and is especially important for
    required call parameters that must be exposed as ONNX inputs.
    Any user-supplied parameter must be a graph input (variable), never a constant/initializer.
    """
    for internal_var, (param_name, param_value) in zip(
        remaining_internal_vars, extra_param_inputs, strict=False
    ):
        # Use the actual parameter name as the internal name for better readability
        internal_name = param_name

        # Handle the mapping in the converter
        if internal_var in sub_converter.var_to_name:
            old_name = sub_converter.var_to_name[internal_var]
            logger.debug(
                f"Replacing generic name '{old_name}' with descriptive name '{internal_name}' for parameter '{param_name}'"
            )

            # Clean up old mappings
            if old_name in sub_converter.name_to_var:
                del sub_converter.name_to_var[old_name]

            # Update mappings for all references to this variable in the graph
            for node in sub_builder.nodes:
                for i, input_name in enumerate(node.input):
                    if input_name == old_name:
                        node.input[i] = internal_name
                        logger.debug(
                            f"Updated node input from '{old_name}' to '{internal_name}'"
                        )

        # Update the converter mappings
        sub_converter.var_to_name[internal_var] = internal_name
        sub_converter.name_to_var[internal_name] = internal_var

        # Register the value info with proper metadata as a graph input (never a constant/initializer)
        # Try to infer shape and dtype from param_value if possible, otherwise default to scalar float32
        try:
            if hasattr(param_value, "shape") and hasattr(param_value, "dtype"):
                shape = tuple(param_value.shape)
                dtype = helper.np_dtype_to_tensor_dtype(param_value.dtype)
            elif isinstance(param_value, (bool, int, float)):
                shape = ()

                if isinstance(param_value, bool):
                    dtype = onnx.TensorProto.BOOL
                elif isinstance(param_value, (int, np.integer)):
                    dtype = onnx.TensorProto.INT32
                else:
                    dtype = onnx.TensorProto.FLOAT
            else:
                shape = ()
                dtype = onnx.TensorProto.FLOAT
        except Exception:
            shape = ()
            dtype = onnx.TensorProto.FLOAT

        sub_builder.register_value_info_metadata(
            internal_name, shape, dtype, origin="function_param_input"
        )
        sub_builder.add_value_info(internal_name, shape, dtype)


def collect_used_param_inputs(sub_builder, parent_builder):
    initializer_names = {i.name for i in parent_builder.initializers}
    used_constants = {
        inp
        for node in sub_builder.nodes
        for inp in node.input
        if inp in initializer_names
    }
    return sorted(used_constants)


def create_function_call(
    unique_node_name,
    input_names,
    param_inputs,
    call_outputs,
    parent_builder,
    display_name,
):
    """Create a function call node that invokes the ONNX function in the parent graph.

    Args:
        unique_node_name: Unique name for the function
        input_names: List of standard input tensor names
        param_inputs: List of parameter input names (constants, weights, etc.)
        call_outputs: List of output tensor names
        parent_builder: The parent ONNX graph builder
        display_name: Human-readable name for debugging
    """
    # Ensure we include all parameter inputs in the final call inputs without duplicates
    # This combines our regular inputs with weight parameters and scalar parameters like 'deterministic'

    # Use a dictionary to track seen inputs
    seen_inputs = {}
    call_inputs = []

    # First add standard input tensor names
    for name in input_names:
        if name not in seen_inputs:
            call_inputs.append(name)
            seen_inputs[name] = True

    # Then add parameter input names, avoiding duplicates
    for name in param_inputs:
        if name not in seen_inputs:
            call_inputs.append(name)
            seen_inputs[name] = True
        else:
            logger.info(
                f"Skipping duplicate input '{name}' in function call to {unique_node_name}"
            )

    parent_builder.add_function_call_node(
        function_name=unique_node_name,
        input_names=call_inputs,
        output_names=call_outputs,
        node_name=unique_node_name,
        user_display_name=display_name,
    )

    logger.debug(f"✅ Added call node for: {unique_node_name}")


def trace_function_body(
    converter,
    orig_fn,
    params,
    example_args,
    unique_node_name,
    parent_builder,
    outer_input_vars_avals,
    extra_param_inputs,
    eqn,
):
    """Trace the function body and set up the function's inputs.

    This function performs the JAX function tracing and registers all inputs
    including regular inputs and parameter inputs.

    Args:
        converter: The parent converter instance
        orig_fn: Original JAX function to trace
        params: Function parameters
        example_args: Example arguments for tracing
        unique_node_name: Unique name for the function
        parent_builder: The parent ONNX graph builder
        outer_input_vars_avals: Input variable information from parent scope
        extra_param_inputs: Extra parameter inputs information
        eqn: Equation from JAX's JaxprTrace

    Returns:
        A tuple of (sub_converter, sub_builder, internal_input_vars)
    """

    sub_converter, sub_builder, params = setup_sub_converter(
        converter, eqn, params, unique_node_name, parent_builder
    )

    # Propagate equation parameters (fix for dual param dicts)
    if isinstance(params, dict) and "onnx_params" in params and "jax_params" in params:
        params_to_use = params["jax_params"]
    else:
        params_to_use = params
    params_to_use = propagate_eqn_parameters(eqn, params_to_use)
    sub_converter.params = params_to_use
    params = params_to_use

    trace_kwargs, example_args = prepare_trace_kwargs_and_example_args(
        params, example_args
    )
    sub_converter.trace_jaxpr(orig_fn, example_args, **trace_kwargs)

    internal_input_vars = sub_converter.jaxpr.invars
    register_function_inputs(
        sub_converter, sub_builder, internal_input_vars, outer_input_vars_avals
    )

    remaining_internal_vars = internal_input_vars[len(outer_input_vars_avals) :]
    rename_and_register_param_inputs(
        sub_converter, sub_builder, remaining_internal_vars, extra_param_inputs
    )

    return sub_converter, sub_builder, internal_input_vars


def map_and_register_outputs(
    unique_node_name, sub_builder, parent_builder, sub_converter, converter, eqn
):
    sub_output_names = [vi.name for vi in sub_builder.outputs]
    logger.debug(f"[⚠️ DEBUG] Subgraph output names: {sub_output_names}")
    logger.debug("[⚠️ DEBUG] Mapping subgraph outputs to top-level ONNX outputs:")

    call_outputs = []
    for i, sub_name in enumerate(sub_output_names):
        var = eqn.outvars[i]

        if sub_name not in sub_builder.value_info_metadata:
            sub_var = sub_converter.name_to_var.get(sub_name)
            if sub_var and hasattr(sub_var, "aval"):
                aval = sub_var.aval
                shape = tuple(aval.shape)
                dtype = helper.np_dtype_to_tensor_dtype(aval.dtype)
                sub_builder.register_value_info_metadata(
                    sub_name, shape, dtype, origin="function_output"
                )
                sub_builder.add_value_info(sub_name, shape, dtype)

        shape_dtype = sub_builder.value_info_metadata.get(sub_name)
        if shape_dtype is None:
            raise RuntimeError(
                f"[❌] Missing metadata for subgraph output '{sub_name}'."
            )

        # Get the shape and dtype from metadata
        shape, dtype = shape_dtype

        # Check for symbolic dimensions in shape
        # Skip ShapedArray creation if there are symbolic dimensions - just preserve var.aval
        if all(isinstance(dim, (int, float)) for dim in shape):
            # Only create ShapedArray for concrete shapes
            var.aval = ShapedArray(shape, helper.tensor_dtype_to_np_dtype(dtype))
        else:
            # For symbolic shapes, preserve dimensions but update the dtype if needed
            if hasattr(var, "aval") and var.aval is not None:
                # If we already have an aval with the right shape, just update dtype if needed
                if var.aval.dtype != helper.tensor_dtype_to_np_dtype(dtype):
                    var.aval = var.aval.update(
                        dtype=helper.tensor_dtype_to_np_dtype(dtype)
                    )
            else:
                # If we don't have an aval, we need to construct one that preserves symbolic dims
                # This is a fallback that might not handle all cases
                logger.warning(
                    f"Creating placeholder aval for var with symbolic dimensions: {shape}"
                )
                # Use the first dimension from the original var if available
                if (
                    hasattr(var, "aval")
                    and hasattr(var.aval, "shape")
                    and var.aval.shape
                ):
                    placeholder_shape = var.aval.shape
                else:
                    # Create a default shape - this is not ideal but better than failing
                    placeholder_shape = (2,) * len(shape)
                var.aval = ShapedArray(
                    placeholder_shape, helper.tensor_dtype_to_np_dtype(dtype)
                )

        parent_output_name = parent_builder.get_unique_name("var")
        converter.var_to_name[var] = parent_output_name
        converter.name_to_var[parent_output_name] = var
        call_outputs.append(parent_output_name)

        # Important: Register shape metadata in parent builder
        # We need to preserve the original shape with symbolic dimensions here
        parent_builder.register_value_info_metadata(parent_output_name, shape, dtype)
        # keep the converter’s symbolic table in sync
        if hasattr(converter, "symbolic_shapes"):
            converter.symbolic_shapes[parent_output_name] = shape
        parent_builder.add_value_info(parent_output_name, shape, dtype)

    return call_outputs


# --- little helper -----------------------------------------------------------


# ──────────────────────────────────────────────────────────────────────────────
# utilities
# ──────────────────────────────────────────────────────────────────────────────


def _base_name(name: str) -> str:
    """Return `name` without the trailing ``_<digits>`` suffix."""
    return re.sub(r"_\d+$", "", name)


def function_handler(
    name: str,
    converter: "Jaxpr2OnnxConverter",
    eqn,
    orig_fn: Callable,
    onnx_params,
    jax_params=None,
    params=None,
):
    """
    Convert a primitive produced by ``@onnx_function``.

    Compared with the previous implementation this version
    **(a)** inlines trivial pass‑through wrappers and
    **(b)** registers every non‑trivial function as a proper
    `FunctionProto`, guaranteeing ORT can resolve the
    op‑type at load time.
    """

    # ------------------------------------------------------------------
    # 1) Boiler‑plate: checks, name generation, input preparation
    # ------------------------------------------------------------------
    if orig_fn is None:
        raise RuntimeError(f"Original function for {name} not recorded.")

    impl_key, unique_node_name, parent_builder = prepare_function_names(
        converter, orig_fn, name
    )

    input_names, example_args, outer_input_avals = resolve_function_inputs(
        converter, eqn, parent_builder
    )

    extra_param_inputs = handle_function_parameters(
        params if params is not None else onnx_params,
        converter,
        eqn,
        parent_builder,
        input_names,
        example_args,
        orig_fn=orig_fn,
    )

    logger.debug(f"Tracing function body for: {unique_node_name}")

    # Use jax_params for tracing if provided, else fallback to onnx_params
    trace_params = jax_params if jax_params is not None else onnx_params
    sub_converter, sub_builder, _ = trace_function_body(
        converter,
        orig_fn,
        trace_params,
        example_args,
        unique_node_name,
        parent_builder,
        outer_input_avals,
        extra_param_inputs,
        eqn,
    )

    # ------------------------------------------------------------------
    # 2)  INLINE  trivial wrappers  (the body contains exactly one node
    #     that merely calls the *real* function)
    # ------------------------------------------------------------------
    inner_nodes = list(sub_builder.nodes)
    if len(inner_nodes) == 1 and _base_name(inner_nodes[0].op_type) == _base_name(
        unique_node_name
    ):
        logger.debug(f"Inlining trivial wrapper '{unique_node_name}' (calls itself).")

        parent_builder._propagate_nested_functions(sub_builder)
        parent_builder.merge_value_info_metadata_from(sub_builder)

        call_outputs = map_and_register_outputs(
            inner_nodes[0].op_type,
            sub_builder,
            parent_builder,
            sub_converter,
            converter,
            eqn,
        )

        param_inputs = collect_used_param_inputs(sub_builder, parent_builder)

        create_function_call(
            inner_nodes[0].op_type,
            input_names,
            param_inputs,
            call_outputs,
            parent_builder,
            name,
        )
        return  # wrapper collapsed – done!

    # ------------------------------------------------------------------
    # 3)  NORMAL path – keep the wrapper as a reusable ONNX function
    # ------------------------------------------------------------------
    param_inputs = collect_used_param_inputs(sub_builder, parent_builder)

    parent_builder.add_function(
        name=unique_node_name,
        sub_builder=sub_builder,
        param_input_names=param_inputs,
        sub_converter=sub_converter,
    )

    parent_builder.merge_value_info_metadata_from(sub_builder)

    call_outputs = map_and_register_outputs(
        unique_node_name,
        sub_builder,
        parent_builder,
        sub_converter,
        converter,
        eqn,
    )

    parent_builder._propagate_nested_functions(sub_builder)

    create_function_call(
        unique_node_name,
        input_names,
        param_inputs,
        call_outputs,
        parent_builder,
        name,
    )



# file: jax2onnx/plugins/examples/onnx_functions/onnx_functions_005.py

import jax.numpy as jnp
from flax import nnx

from jax2onnx.plugin_system import onnx_function, register_example


@onnx_function
class NestedBlock(nnx.Module):

    def __init__(self, num_hiddens, mlp_dim, dropout_rate=0.1, *, rngs: nnx.Rngs):
        self.layers = [
            nnx.Linear(num_hiddens, mlp_dim, rngs=rngs),
            lambda x: nnx.gelu(x, approximate=False),
            nnx.Dropout(rate=0.1, rngs=rngs),
            nnx.Linear(mlp_dim, num_hiddens, rngs=rngs),
            nnx.Dropout(rate=0.1, rngs=rngs),
        ]

    def __call__(self, x: jnp.ndarray, deterministic: bool = True) -> jnp.ndarray:
        for layer in self.layers:
            if isinstance(layer, nnx.Dropout):
                x = layer(x, deterministic=deterministic)
            else:
                x = layer(x)
        return x


@onnx_function
class SuperBlock(nnx.Module):
    def __init__(self):
        rngs = nnx.Rngs(0)
        num_hiddens = 256
        self.layer_norm2 = nnx.LayerNorm(num_hiddens, rngs=rngs)
        self.mlp = NestedBlock(num_hiddens, mlp_dim=512, rngs=rngs)

    def __call__(self, x):
        return self.mlp(self.layer_norm2(x))


register_example(
    component="onnx_functions_005",
    description="nested function plus more components",
    since="v0.4.0",
    context="examples.onnx_functions",
    children=["NestedBlock"],
    testcases=[
        {
            "testcase": "005_nested_function_plus_component",
            "callable": SuperBlock(),
            "input_shapes": [("B", 10, 256)],
            "expected_number_of_function_instances": 2,
        },
    ],
)



import jax.numpy as jnp
import numpy as np
import onnx
import onnxruntime as ort


# Define tile operation
def _my_tile(x):
    repeats_tensor = jnp.array([3, 1, 1], dtype=jnp.int32)
    tile = jnp.tile(x, repeats_tensor)
    return tile


# Prepare input
x = jnp.ones((1, 1, 8))
out = _my_tile(x)
print("JAX output shape:", out.shape)

# Load ONNX model
onnx_model_path = "docs/onnx/primitives/jnp/tile_repeats_tensor.onnx"
model = onnx.load(onnx_model_path)

# Run ONNX model
ort_session = ort.InferenceSession(onnx_model_path)
ort_inputs = {ort_session.get_inputs()[0].name: np.asarray(x)}
out_onnx = ort_session.run(None, ort_inputs)[0]

print("ONNX output shape:", out_onnx.shape)
# compare the outputs
print("Outputs are equal:", np.array_equal(out, out_onnx))



import jax.numpy as jnp
import onnx
from jax2onnx import to_onnx
from jax2onnx.plugin_system import onnx_function
from logging_config import configure_logging


configure_logging()


@onnx_function
def fn(x):
    return jnp.squeeze(x, (-1, -3))


def symbolic_batch_dim_is_preserved():
    # Use abstracted axes with a symbolic name "B"

    # Convert the function to ONNX
    model = to_onnx(fn=lambda x: fn(x), inputs=[(1, "B", 1)])

    onnx.save_model(model, "docs/onnx/test_symbolic_dim_name_in_fn.onnx")

    print(model)


# add main
if __name__ == "__main__":
    symbolic_batch_dim_is_preserved()



# file: jax2onnx/plugins/examples/onnx_functions/onnx_functions_010.py


import jax.numpy as jnp
from flax import nnx

from jax2onnx.plugin_system import onnx_function, register_example


@onnx_function
class FeedForward(nnx.Module):
    def __init__(self, num_hiddens, mlp_dim, dropout_rate=0.1, *, rngs: nnx.Rngs):
        self.layers = [
            nnx.Linear(num_hiddens, mlp_dim, rngs=rngs),
            lambda x: nnx.gelu(x, approximate=False),
            nnx.Dropout(rate=0.1, rngs=rngs),
            nnx.Linear(mlp_dim, num_hiddens, rngs=rngs),
            nnx.Dropout(rate=0.1, rngs=rngs),
        ]

    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        deterministic = True
        for layer in self.layers:
            if isinstance(layer, nnx.Dropout):
                x = layer(x, deterministic=deterministic)
            else:
                x = layer(x)
        return x


@onnx_function
def attention(*args, **kwargs):
    return nnx.dot_product_attention(*args, **kwargs)


@onnx_function
class MultiHeadAttention(nnx.Module):
    def __init__(
        self,
        num_hiddens: int,
        num_heads: int,
        attention_dropout_rate: float = 0.1,
        *,
        rngs: nnx.Rngs,
    ):
        self.attention = nnx.MultiHeadAttention(
            num_heads=num_heads,
            qkv_features=num_hiddens,
            out_features=num_hiddens,
            in_features=num_hiddens,
            attention_fn=lambda *args, **kwargs: attention(*args),
            rngs=rngs,
            decode=False,
        )
        self.dropout = nnx.Dropout(rate=attention_dropout_rate, rngs=rngs)

    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        x = self.attention(x, deterministic=True)
        x = self.dropout(x, deterministic=True)
        return x


@onnx_function
class TransformerBlock(nnx.Module):
    def __init__(
        self,
        num_hiddens: int,
        num_heads: int,
        mlp_dim: int,
        attention_dropout_rate: float = 0.1,
        mlp_dropout_rate: float = 0.1,
        *,
        rngs: nnx.Rngs,
    ):
        self.layer_norm1 = nnx.LayerNorm(num_hiddens, rngs=rngs)
        self.attention = MultiHeadAttention(
            num_hiddens=num_hiddens,
            num_heads=num_heads,
            attention_dropout_rate=attention_dropout_rate,
            rngs=rngs,
        )
        self.layer_norm2 = nnx.LayerNorm(num_hiddens, rngs=rngs)
        self.mlp_block = FeedForward(num_hiddens, mlp_dim, mlp_dropout_rate, rngs=rngs)

    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        r = self.layer_norm1(x)
        r = self.attention(r)
        x = x + r
        r = self.layer_norm2(x)
        return x + self.mlp_block(r)


@onnx_function
class TransformerStack(nnx.Module):
    def __init__(
        self,
        num_hiddens: int,
        num_heads: int,
        mlp_dim: int,
        num_layers: int,
        attention_dropout_rate: float = 0.1,
        mlp_dropout_rate: float = 0.1,
        *,
        rngs: nnx.Rngs,
    ):
        self.blocks = [
            TransformerBlock(
                num_hiddens,
                num_heads,
                mlp_dim,
                attention_dropout_rate,
                mlp_dropout_rate,
                rngs=rngs,
            )
            for _ in range(num_layers)
        ]

    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        for block in self.blocks:
            x = block(x)
        return x


register_example(
    component="onnx_functions_010",
    description="transformer stack",
    since="v0.4.0",
    context="examples.onnx_functions",
    children=["TransformerBlock"],
    testcases=[
        {
            "testcase": "010_transformer_stack",
            "callable": TransformerStack(
                num_hiddens=256,
                num_heads=8,
                mlp_dim=512,
                num_layers=6,
                attention_dropout_rate=0.1,
                mlp_dropout_rate=0.1,
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [("B", 10, 256)],
            "expected_number_of_function_instances": 25,
        },
    ],
)



"""
Plugin for handling the JAX device_put primitive.

This plugin converts JAX's device_put primitive to appropriate ONNX operations.
"""

from typing import TYPE_CHECKING

import jax
import numpy as np
from jax.extend import core as extend_core

from jax2onnx.plugin_system import PrimitiveLeafPlugin, register_primitive

if TYPE_CHECKING:
    from jax2onnx.converter.jaxpr_converter import Jaxpr2OnnxConverter


@register_primitive(
    jaxpr_primitive=jax.lax.device_put_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.device_put.html#jax.device_put",
    onnx=[
        {
            "component": "Identity",
            "doc": "https://onnx.ai/onnx/operators/onnx__Identity.html",
        }
    ],
    since="v0.4.0",
    context="primitives.lax",
    component="device_put",
    testcases=[
        {
            "testcase": "device_put_array",
            "callable": lambda x: jax.device_put(x),
            "input_shapes": [(3, 4)],
        },
        {
            "testcase": "device_put_scalar",
            "callable": lambda: jax.device_put(42),
            "input_shapes": [],
        },
    ],
)
class DevicePutPlugin(PrimitiveLeafPlugin):
    """Plugin for converting jax.lax.device_put to appropriate ONNX operations."""

    def to_onnx(
        self, converter: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params
    ):
        """
        Convert jax.lax.device_put to ONNX operations.

        For constants, creates a constant node.
        For variables, creates an Identity node.

        Arguments:
            converter: The Jaxpr2OnnxConverter instance
            node_inputs: Input variables to the primitive
            node_outputs: Output variables from the primitive
            params: Parameters for the primitive
        """
        inp = node_inputs[0]
        out = node_outputs[0]

        if isinstance(inp, extend_core.Literal):
            # Handle conversion of literal values
            val = inp.val
            np_val = np.array(val)

            # Check output type and ensure we match it
            output_aval = out.aval
            output_dtype = output_aval.dtype

            # Convert value to match expected output dtype
            if np_val.dtype != output_dtype:
                np_val = np_val.astype(output_dtype)

            # Get tensor name for the constant
            tensor_name = converter.get_unique_name("const")

            # Use add_initializer to add the constant to the ONNX graph
            # This will handle data type conversion and initialization

            data_type = converter.builder._numpy_dtype_to_onnx(np_val.dtype)
            converter.builder.add_initializer(
                tensor_name, np_val.flatten().tolist(), data_type, dims=np_val.shape
            )

            output_name = converter.get_name(out)
            node = converter.builder.create_node(
                "Identity",
                [tensor_name],
                [output_name],
                name=converter.get_unique_name("device_put"),
            )
            converter.add_node(node)
        else:
            # For non-literal inputs, simply pass through with Identity
            input_names = [converter.get_name(inp) for inp in node_inputs]
            output_names = [converter.get_name(out) for out in node_outputs]
            if not output_names:
                return

            node = converter.builder.create_node(
                "Identity",
                input_names,
                output_names,
                name=converter.get_unique_name(f"identity_{jax.lax.device_put_p.name}"),
            )
            converter.add_node(node)



from typing import TYPE_CHECKING

from flax import nnx
from jax import core
from jax.extend.core import Primitive
from onnx import helper

from jax2onnx.plugin_system import PrimitiveLeafPlugin, register_primitive

if TYPE_CHECKING:
    from jax2onnx.converter.jaxpr_converter import Jaxpr2OnnxConverter

# Define the ReLU primitive
nnx.relu_p = Primitive("nnx.relu")
nnx.relu_p.multiple_results = False  # Correct initialization


@register_primitive(
    jaxpr_primitive=nnx.relu_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.nn.relu.html",
    onnx=[
        {
            "component": "Relu",
            "doc": "https://onnx.ai/onnx/operators/onnx__Relu.html",
        }
    ],
    since="v0.1.0",
    context="primitives.nnx",
    component="relu",
    testcases=[
        {
            "testcase": "relu_1d",
            "callable": lambda x: nnx.relu(x),
            "input_shapes": [(3,)],
        },
        {
            "testcase": "relu_4d",
            "callable": lambda x: nnx.relu(x),
            "input_shapes": [("B", 28, 28, 32)],
        },
    ],
)
class ReluPlugin(PrimitiveLeafPlugin):
    """
    Plugin for converting flax.nnx.relu to ONNX.
    """

    @staticmethod
    def abstract_eval(x):
        """Abstract evaluation function for ReLU."""
        return core.ShapedArray(x.shape, x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of ReLU to ONNX format."""
        input_var = node_inputs[0]
        output_var = node_outputs[0]

        input_name = s.get_name(input_var)
        output_name = s.get_name(output_var)

        relu_node = helper.make_node(
            "Relu",
            inputs=[input_name],
            outputs=[output_name],
            name=s.get_unique_name("relu"),
        )
        s.add_node(relu_node)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for ReLU."""

        def patched_relu(x):
            return nnx.relu_p.bind(x)

        return patched_relu

    @staticmethod
    def patch_info():
        """Provides patching information for ReLU."""
        return {
            "patch_targets": [nnx],
            "patch_function": lambda _: ReluPlugin.get_monkey_patch(),
            "target_attribute": "relu",
        }


# Register abstract evaluation function
nnx.relu_p.def_abstract_eval(ReluPlugin.abstract_eval)



import os

import onnx
from onnx import TensorProto, helper

# Define a unique domain for our custom functions
CUSTOM_DOMAIN = "my.custom.domain.embedded"
OPSET_VERSION = 14  # Choose an appropriate opset version


# === Helper Function to Create Constant Node for Float Scalar ===
def make_float_scalar_constant_node(name, output_name, value):
    """Creates an ONNX Constant node for a scalar float value."""
    return helper.make_node(
        "Constant",
        inputs=[],
        outputs=[output_name],
        name=name,
        value=helper.make_tensor(
            name=output_name + "_value",  # Tensor name inside attribute
            data_type=TensorProto.FLOAT,
            dims=[],  # Scalar
            vals=[float(value)],
        ),
    )


# === 1. Define FunctionProto 1: ScaleAndAdd_A ===
# Function: Y_out = (X_in * 2.0) + 1.0
# Constants 2.0 and 1.0 are embedded using Constant nodes

func_A_input_names = ["X_in_A"]
func_A_output_names = ["Y_out_A"]

# Define Constant nodes *inside* the function for the scalar values
func_A_node_const_scale = make_float_scalar_constant_node(
    name="FuncA_Const_Scale", output_name="scale_A_local_val", value=2.0
)
func_A_node_const_bias = make_float_scalar_constant_node(
    name="FuncA_Const_Bias", output_name="bias_A_local_val", value=1.0
)

# Define computation nodes using the output names of the Constant nodes
func_A_node_mul = helper.make_node(
    "Mul",
    inputs=["X_in_A", "scale_A_local_val"],  # Use Constant node output
    outputs=["mul_result_A"],
    name="FuncA_Node_Mul",
)
func_A_node_add = helper.make_node(
    "Add",
    inputs=["mul_result_A", "bias_A_local_val"],  # Use Constant node output
    outputs=["Y_out_A"],
    name="FuncA_Node_Add",
)

# Create the FunctionProto for ScaleAndAdd_A
# Include the Constant nodes in the 'nodes' list
scale_and_add_func_A = helper.make_function(
    domain=CUSTOM_DOMAIN,
    fname="ScaleAndAdd_A",
    inputs=func_A_input_names,
    outputs=func_A_output_names,
    nodes=[
        func_A_node_const_scale,  # Constant node comes first
        func_A_node_const_bias,  # Constant node comes first
        func_A_node_mul,
        func_A_node_add,
    ],
    # NO 'initializer' argument here
    opset_imports=[helper.make_opsetid("", OPSET_VERSION)],
)

# === 2. Define FunctionProto 2: ScaleAndAdd_B ===
# Function: Y_out = (X_in * 3.0) + 5.0
# Constants 3.0 and 5.0 are embedded using Constant nodes

func_B_input_names = ["X_in_B"]
func_B_output_names = ["Y_out_B"]

# Define Constant nodes *inside* this function
func_B_node_const_scale = make_float_scalar_constant_node(
    name="FuncB_Const_Scale",
    output_name="scale_B_local_val",
    value=3.0,  # Different value
)
func_B_node_const_bias = make_float_scalar_constant_node(
    name="FuncB_Const_Bias",
    output_name="bias_B_local_val",
    value=5.0,  # Different value
)

# Define computation nodes using the output names of the Constant nodes
func_B_node_mul = helper.make_node(
    "Mul",
    inputs=["X_in_B", "scale_B_local_val"],  # Use Constant node output
    outputs=["mul_result_B"],
    name="FuncB_Node_Mul",
)
func_B_node_add = helper.make_node(
    "Add",
    inputs=["mul_result_B", "bias_B_local_val"],  # Use Constant node output
    outputs=["Y_out_B"],
    name="FuncB_Node_Add",
)

# Create the FunctionProto for ScaleAndAdd_B
scale_and_add_func_B = helper.make_function(
    domain=CUSTOM_DOMAIN,
    fname="ScaleAndAdd_B",
    inputs=func_B_input_names,
    outputs=func_B_output_names,
    nodes=[
        func_B_node_const_scale,
        func_B_node_const_bias,
        func_B_node_mul,
        func_B_node_add,
    ],
    # NO 'initializer' argument here
    opset_imports=[helper.make_opsetid("", OPSET_VERSION)],
)


# === 3. Define the Main GraphProto ===
# (This part remains the same as the previous attempt)

graph_input_1 = helper.make_tensor_value_info("Input_1", TensorProto.FLOAT, None)
graph_input_2 = helper.make_tensor_value_info("Input_2", TensorProto.FLOAT, None)
graph_output_1 = helper.make_tensor_value_info("Output_1", TensorProto.FLOAT, None)
graph_output_2 = helper.make_tensor_value_info("Output_2", TensorProto.FLOAT, None)

graph_node_1 = helper.make_node(
    op_type="ScaleAndAdd_A",
    inputs=["Input_1"],
    outputs=["Output_1"],
    name="FunctionCall_A",
    domain=CUSTOM_DOMAIN,
)
graph_node_2 = helper.make_node(
    op_type="ScaleAndAdd_B",
    inputs=["Input_2"],
    outputs=["Output_2"],
    name="FunctionCall_B",
    domain=CUSTOM_DOMAIN,
)

main_graph = helper.make_graph(
    nodes=[graph_node_1, graph_node_2],
    name="MainGraph_EmbeddedConstants",
    inputs=[graph_input_1, graph_input_2],
    outputs=[graph_output_1, graph_output_2],
    initializer=[],  # No initializers needed here for scale/bias
)

# === 4. Assemble and Save the ModelProto ===
# (This part remains the same)

opset_imports = [
    helper.make_opsetid("", OPSET_VERSION),
    helper.make_opsetid(CUSTOM_DOMAIN, 1),
]

model = helper.make_model(
    main_graph,
    functions=[
        scale_and_add_func_A,
        scale_and_add_func_B,
    ],  # Include both function defs
    opset_imports=opset_imports,
    producer_name="onnx-embedded-constants-example-v2",
)

try:
    onnx.checker.check_model(model)
    print("ONNX model check passed.")
except onnx.checker.ValidationError as e:
    print(f"ONNX model check failed: {e}")

output_dir = "./onnx_output"
os.makedirs(output_dir, exist_ok=True)
# Use a different filename to avoid confusion
output_path = os.path.join(output_dir, "embedded_constants_corrected_example.onnx")

onnx.save(model, output_path)

print(f"✅ Corrected ONNX model saved to: {output_path}")
print("This version uses Constant nodes inside the functions to embed parameters.")



# file: jax2onnx/plugins/flax/nnx/batch_norm.py

"""
Batch Norm Plugin for JAX to ONNX conversion.

This plugin enables conversion of flax.nnx.BatchNorm layers (in inference mode)
to ONNX format. It transforms JAX’s batch_norm operations into an ONNX
BatchNormalization operator with necessary Transpose operations for
NHWC/NLC to NCHW/NCL conversion and adds required shape information.

The conversion process involves:
  1. Defining a JAX primitive for BatchNorm's inference behavior.
  2. Providing an abstract evaluation for JAX's tracing system.
  3. Handling input shape, transpositions (2D, 3D, 4D), and adding ONNX shape info.
  4. Converting the operation to ONNX using BatchNormalization and Transpose nodes.
  5. Monkey-patching BatchNorm.__call__ to redirect calls to our primitive,
     ensuring inference parameters (running mean/var) and default scale/bias
     are used.
"""

from typing import TYPE_CHECKING

import numpy as np
from flax import nnx
from jax.extend.core import Primitive
from onnx import helper

from jax2onnx.plugin_system import PrimitiveLeafPlugin, register_primitive

if TYPE_CHECKING:
    from jax2onnx.converter.jaxpr_converter import Jaxpr2OnnxConverter

# Define the BatchNorm primitive
nnx.batch_norm_p = Primitive("nnx.batch_norm")
nnx.batch_norm_p.multiple_results = False  # Correctly set at initialization


@register_primitive(
    jaxpr_primitive=nnx.batch_norm_p.name,
    jax_doc="https://flax.readthedocs.io/en/latest/api_reference/flax.nnx/nn/normalization.html#flax.nnx.BatchNorm",
    onnx=[
        {
            "component": "BatchNormalization",
            "doc": "https://onnx.ai/onnx/operators/onnx__BatchNormalization.html",
        }
    ],
    since="v0.1.0",
    context="primitives.nnx",
    component="batch_norm",
    testcases=[
        {
            "testcase": "batch_norm_simple",
            "callable": nnx.BatchNorm(
                num_features=1, use_running_average=True, epsilon=1e-5, rngs=nnx.Rngs(0)
            ),
            "input_shapes": [("B", 1)],
        },
        {
            "testcase": "batch_norm_2d",
            "callable": nnx.BatchNorm(
                num_features=8, use_running_average=True, epsilon=1e-5, rngs=nnx.Rngs(0)
            ),
            "input_shapes": [("B", 8)],
        },
        {
            "testcase": "batch_norm_2d_use_bias_false",
            "callable": nnx.BatchNorm(
                num_features=8,
                use_running_average=True,
                use_bias=False,
                epsilon=1e-5,
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [("B", 8)],
        },
        {
            "testcase": "batch_norm_2d_use_scale_false",
            "callable": nnx.BatchNorm(
                num_features=8,
                use_running_average=True,
                use_scale=False,
                epsilon=1e-5,
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [("B", 8)],
        },
        {
            "testcase": "batch_norm_4d",
            "callable": nnx.BatchNorm(
                num_features=3, use_running_average=True, epsilon=1e-5, rngs=nnx.Rngs(0)
            ),
            "input_shapes": [("B", 4, 4, 3)],
        },
        {
            "testcase": "batch_norm_4d_use_bias_false",
            "callable": nnx.BatchNorm(
                num_features=3,
                use_running_average=True,
                use_bias=False,
                epsilon=1e-5,
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [("B", 4, 4, 3)],
        },
        {
            "testcase": "batch_norm_4d_use_scale_false",
            "callable": nnx.BatchNorm(
                num_features=3,
                use_running_average=True,
                use_scale=False,
                epsilon=1e-5,
                rngs=nnx.Rngs(0),
            ),
            "input_shapes": [("B", 4, 4, 3)],
        },
        {
            "testcase": "batch_norm_minimal",
            "callable": nnx.BatchNorm(
                num_features=1, use_running_average=True, epsilon=1e-5, rngs=nnx.Rngs(0)
            ),
            "input_shapes": [(1, 1)],
            "input_values": np.array([[0.0]], dtype=np.float32),
        },
    ],
)
class BatchNormPlugin(PrimitiveLeafPlugin):
    """
    Plugin for converting flax.nnx.BatchNorm to ONNX.
    """

    @staticmethod
    def abstract_eval(x, **kwargs):
        """Abstract evaluation function for BatchNorm."""
        # Use update instead of creating a new ShapedArray to avoid issues with unhashable tracers
        return x.update(shape=x.shape, dtype=x.dtype, weak_type=False)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of BatchNorm to ONNX format."""
        input_name = s.get_name(node_inputs[0])
        output_name = s.get_name(node_outputs[0])

        params.get("use_running_average", False)
        params.get("axis", -1)
        params.get("axis_index_groups", None)
        params.get("axis_name", None)
        use_bias = params.get("use_bias", True)
        params.get("use_fast_variance", False)
        use_scale = params.get("use_scale", True)
        dtype = getattr(node_inputs[0].aval, "dtype", np.float32)
        params.get("num_features", None)
        bias = params.get("bias", None)
        scale = params.get("scale", None)
        var = params.get("var", None)
        mean = params.get("mean", None)
        epsilon = params.get("epsilon")
        momentum = params.get("momentum", 0.9)
        params.get("use_running_average", False)

        # --- handle NHWC/NCHW transpose and param shapes ---
        input_shape = node_inputs[0].aval.shape
        input_rank = len(input_shape)
        channel_axis = input_rank - 1  # NHWC/NLC: last axis
        C = input_shape[channel_axis]

        # For NHWC->NCHW and NLC->NCL, channel order is preserved, so no reordering is needed.
        # Just ensure parameters are 1D (C,) and ONNX initializers.
        def ensure_onnx_param(param, name, default_value):
            arr = (
                np.array(param, dtype=dtype)
                if param is not None
                else np.full((C,), default_value, dtype=dtype)
            )
            arr = arr.reshape((C,))
            if arr.shape != (C,):
                raise ValueError(
                    f"BatchNorm param '{name}' must be shape ({C},), got {arr.shape}"
                )
            return s.builder.get_constant_name(arr)

        scale_name = (
            ensure_onnx_param(scale, "scale", 1.0)
            if use_scale
            else s.builder.get_constant_name(np.ones((C,), dtype=dtype))
        )
        bias_name = (
            ensure_onnx_param(bias, "bias", 0.0)
            if use_bias
            else s.builder.get_constant_name(np.zeros((C,), dtype=dtype))
        )
        mean_name = ensure_onnx_param(mean, "mean", 0.0)
        variance_name = ensure_onnx_param(var, "var", 1.0)

        # Insert transpose if needed
        pre_perm = post_perm = None
        if input_rank == 4:
            pre_perm = [0, 3, 1, 2]
            post_perm = [0, 2, 3, 1]
        elif input_rank == 3:
            pre_perm = [0, 2, 1]
            post_perm = [0, 2, 1]

        bn_input = input_name
        if pre_perm:
            transposed_input = s.get_unique_name("bn_pre_transpose")
            pre_transpose_node = helper.make_node(
                "Transpose",
                inputs=[input_name],
                outputs=[transposed_input],
                name=s.get_unique_name("bn_transpose_pre"),
                perm=pre_perm,
            )
            s.add_node(pre_transpose_node)
            transposed_shape = tuple(input_shape[i] for i in pre_perm)
            s.add_shape_info(transposed_input, transposed_shape, dtype)
            bn_input = transposed_input

        bn_output = output_name
        if post_perm:
            bn_output = s.get_unique_name("bn_output")

        inputs = [bn_input, scale_name, bias_name, mean_name, variance_name]

        # Set ONNX momentum attribute to match JAX/Flax value
        bn_node = helper.make_node(
            "BatchNormalization",
            inputs=inputs,
            outputs=[bn_output],
            name=s.get_unique_name("batch_norm"),
            epsilon=epsilon,
            momentum=momentum,  # Explicitly set momentum to match params
        )
        s.add_node(bn_node)
        if post_perm:
            s.add_shape_info(bn_output, transposed_shape, dtype)
        if post_perm:
            post_transpose_node = helper.make_node(
                "Transpose",
                inputs=[bn_output],
                outputs=[output_name],
                name=s.get_unique_name("bn_transpose_post"),
                perm=post_perm,
            )
            s.add_node(post_transpose_node)
            s.add_shape_info(output_name, input_shape, dtype)

    @staticmethod
    def _batch_norm(
        x,
        use_running_average,
        axis,
        axis_index_groups,
        axis_name,
        bias,
        dtype,
        epsilon,
        mean,
        momentum,
        num_features,
        scale,
        use_bias,
        use_fast_variance,
        use_scale,
        var,
    ):
        """Defines the primitive binding for BatchNorm."""
        return nnx.batch_norm_p.bind(
            x,
            use_running_average=use_running_average,
            axis=axis,
            axis_index_groups=axis_index_groups,
            axis_name=axis_name,
            bias=bias,
            dtype=dtype,
            epsilon=epsilon,
            mean=mean,
            momentum=momentum,
            num_features=num_features,
            scale=scale,
            use_bias=use_bias,
            use_fast_variance=use_fast_variance,
            use_scale=use_scale,
            var=var,
        )

    @staticmethod
    def get_monkey_patch():
        """Returns a patched version of BatchNorm's call method."""

        def patched_batch_norm_call(self, x, use_running_average=None, *, mask=None):
            return BatchNormPlugin._batch_norm(
                x,
                self.use_running_average,
                self.axis,
                self.axis_index_groups,
                self.axis_name,
                self.bias.value if getattr(self, "bias", None) is not None else None,
                self.dtype,
                self.epsilon,
                self.mean.value,
                self.momentum,
                self.num_features,
                self.scale.value if getattr(self, "scale", None) is not None else None,
                self.use_bias,
                self.use_fast_variance,
                self.use_scale,
                self.var.value,
            )

        return patched_batch_norm_call

    @staticmethod
    def patch_info():
        """Provides patching information for BatchNorm."""
        return {
            "patch_targets": [nnx.BatchNorm],
            "patch_function": lambda _: BatchNormPlugin.get_monkey_patch(),
            "target_attribute": "__call__",
        }


# Register abstract evaluation function
nnx.batch_norm_p.def_abstract_eval(BatchNormPlugin.abstract_eval)



"""
Demo for working with JAX symbolic shapes in export and concat operations.
Demonstrates concatenating two tensors with dynamic batch dimension.
"""

import numpy as np
import jax
from jax import export
from jax import numpy as jnp


def f(a, b):  # a: f32[B, 1, 8], b: f32[B, 10, 8]
    """Concatenate two tensors along axis 1."""
    return jnp.concat((a, b), axis=1)


def demo_symbolic_shapes():
    """Demonstrate symbolic shapes with JAX export for concatenation."""
    # Construct symbolic dimension variables
    B_tuple = export.symbolic_shape("B")
    B = B_tuple[0]  # Extract the single dimension from the tuple
    print(f"Created symbolic dimension: B={B}")

    # Use symbolic dimensions to construct shapes
    a_shape = (B, 1, 8)
    b_shape = (B, 10, 8)
    print(f"Symbolic input shapes: a={a_shape}, b={b_shape}")

    # Export with symbolic shapes
    exported = export.export(jax.jit(f))(
        jax.ShapeDtypeStruct(a_shape, jnp.float32),
        jax.ShapeDtypeStruct(b_shape, jnp.float32),
    )

    print("Export complete!")
    print(f"Input shapes: {exported.in_avals}")
    print(f"Output shapes: {exported.out_avals}")

    # Call with concrete shapes (with B=3), without re-tracing
    batch_size = 3
    a_input = np.ones((batch_size, 1, 8), dtype=np.float32)
    b_input = np.ones((batch_size, 10, 8), dtype=np.float32) * 2

    result = exported.call(a_input, b_input)

    print(f"\nConcrete input shapes: a={a_input.shape}, b={b_input.shape}")
    print(f"Concrete output shape: {result.shape}")
    print(f"Output data (first sample, first few values): {result[0, 0:5, 0]}")

    # Try another batch size (B=5)
    batch_size = 5
    a_input = np.ones((batch_size, 1, 8), dtype=np.float32) * 3
    b_input = np.ones((batch_size, 10, 8), dtype=np.float32) * 4

    another_result = exported.call(a_input, b_input)

    print(f"\nAnother batch size (B={batch_size}):")
    print(f"Output shape: {another_result.shape}")
    print(f"Output data (first sample, first few values): {another_result[0, 0:5, 0]}")


if __name__ == "__main__":
    print("JAX Symbolic Shapes with Dynamic Batch Concatenation Demo")
    print("--------------------------------------------------------")
    demo_symbolic_shapes()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Callable, Sequence

import jax
import numpy as np
from jax import lax
from jax.extend.core import Primitive, Var
from onnx import helper

from jax2onnx.converter.onnx_builder import OnnxBuilder
from jax2onnx.plugin_system import PrimitiveLeafPlugin, register_primitive

if TYPE_CHECKING:
    from jax2onnx.converter.jaxpr_converter import Jaxpr2OnnxConverter

"""
jax2onnx plugin: lax.cond → ONNX If   (single tensor operand / result)
"""

logger = logging.getLogger("jax2onnx.plugins.jax.lax.cond")

cond_p = Primitive("cond")
cond_p.multiple_results = True


@register_primitive(
    jaxpr_primitive=cond_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.cond.html",
    onnx=[{"component": "If", "doc": "https://onnx.ai/onnx/operators/onnx__If.html"}],
    since="v0.5.1",
    context="primitives.lax",
    component="cond",
    testcases=[
        {
            "testcase": "cond_scalar",
            "callable": lambda: lax.cond(
                True,
                lambda x: x + 1,
                lambda x: x - 1,
                np.int32(3),
            ),
            "input_shapes": [],
            "expected_output_shapes": [()],
        }
    ],
)
class CondPlugin(PrimitiveLeafPlugin):
    _ORIG_COND: Callable | None = None

    # Only one output – shapes must match between branches
    @staticmethod
    def abstract_eval(pred_aval, *op_aval, true_jaxpr, **__):
        return tuple(op_aval)

    # ------------------------------------------------------------------ #
    # ONNX lowering
    # ------------------------------------------------------------------ #
    def to_onnx(
        self,
        s: "Jaxpr2OnnxConverter",
        node_inputs: Sequence[Var],  # [pred, operand]
        node_outputs: Sequence[Var],
        params: dict[str, Any],
    ):
        pred_var, op_var = node_inputs
        out_var = node_outputs[0]
        pred_name = s.get_name(pred_var)
        op_name = s.get_name(op_var)
        out_name = s.get_name(out_var)

        true_closed = params["true_jaxpr"]
        false_closed = params["false_jaxpr"]

        def _make_branch(name_prefix: str, closed):
            bb = OnnxBuilder(
                name_generator=s.builder.name_generator,
                opset=s.builder.opset,
                model_name=s.builder.get_unique_name(name_prefix),
            )
            bb.initializers = s.builder.initializers  # <- explicitly link initializers!
            bb.var_to_symbol_map = s.builder.var_to_symbol_map
            bc = s.__class__(bb)

            # Map operand directly to outer‑scope tensor (no graph input!)
            bc.var_to_name[closed.jaxpr.invars[0]] = op_name
            # Closed‑over consts
            for cv, cval in zip(closed.jaxpr.constvars, closed.consts):
                bc.var_to_name[cv] = bc.get_constant_name(cval)

            bc._process_jaxpr(closed.jaxpr, closed.consts)
            bb.outputs.clear()
            branch_out = bc.get_name(closed.jaxpr.outvars[0])
            bb.add_output(branch_out, out_var.aval.shape, out_var.aval.dtype)

            # Register operand's shape/dtype as value_info (optional but tidy)
            bb.add_value_info(op_name, op_var.aval.shape, op_var.aval.dtype)

            bb.inputs.clear()  # ← critical: no formal inputs for If‑branch
            return bb.create_graph(bb.model_name, is_subgraph=True)

        then_graph = _make_branch("then", true_closed)
        else_graph = _make_branch("else", false_closed)

        if_node = helper.make_node(
            "If",
            inputs=[pred_name],
            outputs=[out_name],
            then_branch=then_graph,
            else_branch=else_graph,
            name=s.get_unique_name("cond"),
        )
        s.add_node(if_node)
        s.add_shape_info(out_name, out_var.aval.shape, out_var.aval.dtype)

    # ------------------------------------------------------------------ #
    # Monkey‑patch
    # ------------------------------------------------------------------ #
    @staticmethod
    def _cond_binding(pred, true_fun, false_fun, operand):
        true_closed = jax.make_jaxpr(true_fun)(operand)
        false_closed = jax.make_jaxpr(false_fun)(operand)
        flat, tree = jax.tree_util.tree_flatten(operand)
        res = cond_p.bind(pred, *flat, true_jaxpr=true_closed, false_jaxpr=false_closed)
        return jax.tree_util.tree_unflatten(tree, res)

    @staticmethod
    def get_monkey_patch(orig_fn):
        if CondPlugin._ORIG_COND is None:
            CondPlugin._ORIG_COND = orig_fn

        def patched(pred, true_fun, false_fun, operand):
            return CondPlugin._cond_binding(pred, true_fun, false_fun, operand)

        return patched

    @staticmethod
    def patch_info():
        return {
            "patch_targets": [lax],
            "target_attribute": "cond",
            "patch_function": CondPlugin.get_monkey_patch,
        }


cond_p.def_abstract_eval(CondPlugin.abstract_eval)



# file: jax2onnx/plugins/flax/nnx/dropout.py


from typing import TYPE_CHECKING
import jax
import numpy as np
import onnx
from flax import nnx
from jax.core import ShapedArray
from jax.extend.core import Literal, Primitive
from onnx import helper

from jax2onnx.plugin_system import PrimitiveLeafPlugin, register_primitive
import logging

if TYPE_CHECKING:
    from jax2onnx.converter.jaxpr_converter import Jaxpr2OnnxConverter

# Define the new primitive for dropout.
nnx.dropout_p = Primitive("nnx.dropout")
nnx.dropout_p.multiple_results = False  # Single output


@register_primitive(
    jaxpr_primitive=nnx.dropout_p.name,
    jax_doc="https://flax.readthedocs.io/en/latest/api_reference/flax.nnx/nn/stochastic.html#flax.nnx.Dropout",
    onnx=[
        {
            "component": "Dropout",
            "doc": "https://onnx.ai/onnx/operators/onnx__Dropout.html",
        }
    ],
    since="v0.1.0",
    context="primitives.nnx",
    component="dropout",
    testcases=[
        {
            "testcase": "dropout_init_params",
            "callable": nnx.Dropout(rate=0.5, deterministic=True, rngs=nnx.Rngs(5)),
            "input_shapes": [("B", 10)],
        },
        {
            "testcase": "dropout_call_params",
            "callable": nnx.Dropout(rate=0.5, deterministic=False, rngs=nnx.Rngs(5)),
            "input_shapes": [("B", 10)],
            "input_params": {
                "deterministic": True,
            },
        },
    ],
)
class DropoutPlugin(PrimitiveLeafPlugin):
    """
    Plugin for converting flax.nnx.Dropout to ONNX.
    Supports static and dynamic (call-time) 'deterministic'.
    """

    @staticmethod
    def abstract_eval(x, deterministic, *, rate):
        """Abstract evaluation function for dropout."""
        return ShapedArray(x.shape, x.dtype)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        x_name = s.get_name(node_inputs[0])
        det_input = node_inputs[1]
        output_name = s.get_name(node_outputs[0])

        logging.debug("[DEBUG] Dropout to_onnx called")
        logging.debug(f"[DEBUG] Input tensor name: {x_name}")
        logging.debug(f"[DEBUG] Deterministic input: {det_input}")
        logging.debug(f"[DEBUG] Output name: {output_name}")

        det_name = "deterministic"
        # if det_input is a Variable then change_var_name

        if isinstance(det_input, jax._src.core.Var):
            s.change_var_name(det_input, det_name)

        # Static parameter: rate
        rate = params.get("rate", 0.0)
        logging.debug(f"[DEBUG] Dropout rate: {rate}")
        ratio_tensor = np.array(rate, dtype=np.float32)
        ratio_name = s.builder.get_constant_name(ratio_tensor)

        # Handle deterministic: static or dynamic
        if isinstance(det_input, Literal):
            training_mode = not bool(det_input.val)
            logging.debug(
                f"[DEBUG] Static deterministic value: {det_input.val} → training_mode: {training_mode}"
            )
            training_tensor = np.array(training_mode, dtype=bool)
            training_mode_name = s.builder.get_constant_name(training_tensor)

            # Add value_info for the training mode tensor
            s.builder.add_value_info(
                training_mode_name, shape=(), dtype=onnx.TensorProto.BOOL
            )
            logging.debug(
                f"[DEBUG] Added value_info for training_mode: {training_mode_name}"
            )
        else:
            logging.debug("[DEBUG] Dynamic deterministic input detected")
            det_aval = det_input.aval
            det_shape = det_aval.shape
            det_dtype_enum = onnx.TensorProto.BOOL
            # Register the input with the correct name and type
            s.builder.register_value_info_metadata(
                det_name, shape=det_shape, dtype=det_dtype_enum
            )
            s.builder.add_value_info(det_name, shape=det_shape, dtype=det_dtype_enum)
            flipped_name = s.get_unique_name("training_mode")
            not_node = helper.make_node(
                "Not",
                inputs=[det_name],
                outputs=[flipped_name],
                name=s.get_unique_name("not_deterministic"),
            )
            s.add_node(not_node)
            # Add value_info for the flipped training_mode variable
            s.builder.add_value_info(
                flipped_name, shape=det_shape, dtype=det_dtype_enum
            )
            logging.debug(
                f"[DEBUG] Added NOT node to invert deterministic: input={det_name}, output={flipped_name}"
            )
            logging.debug(f"[DEBUG] Added value_info for training_mode: {flipped_name}")
            training_mode_name = flipped_name

        # ONNX Dropout node
        dropout_inputs = [x_name, ratio_name, training_mode_name]
        dropout_node = helper.make_node(
            "Dropout",
            inputs=dropout_inputs,
            outputs=[output_name],
            name=s.get_unique_name("Dropout"),
        )
        s.add_node(dropout_node)
        logging.debug(
            f"[DEBUG] Added Dropout node with inputs: {dropout_inputs}, output: {output_name}"
        )

    @staticmethod
    def _dropout(x, deterministic, rate):
        """Defines the primitive binding for dropout."""
        return nnx.dropout_p.bind(x, deterministic, rate=rate)

    @staticmethod
    def get_monkey_patch():
        """Returns a patched version of Dropout's __call__ method."""

        def patched_dropout_call(self, x, deterministic=None):
            det = deterministic if deterministic is not None else self.deterministic
            return DropoutPlugin._dropout(x, det, self.rate)

        return patched_dropout_call

    @staticmethod
    def patch_info():
        """Provides patching information for dropout."""
        return {
            "patch_targets": [nnx.Dropout],
            "patch_function": lambda _: DropoutPlugin.get_monkey_patch(),
            "target_attribute": "__call__",
        }


# Register abstract evaluation function
nnx.dropout_p.def_abstract_eval(DropoutPlugin.abstract_eval)



# file: jax2onnx/plugins/jax/numpy/reshape.py

from collections.abc import Sequence
from typing import TYPE_CHECKING, Callable, Any

from jax import core
from jax import numpy as jnp
import jax
from jax.extend.core import Primitive

# Use the internal name with an alias for convenience
from jax._src.export.shape_poly import _DimExpr as DimExpr

from onnx import helper, TensorProto

from jax2onnx.plugin_system import PrimitiveLeafPlugin, register_primitive

if TYPE_CHECKING:
    from jax2onnx.converter.jaxpr_converter import Jaxpr2OnnxConverter

import numpy as np

# Define the reshape primitive
reshape_p = Primitive("reshape")
reshape_p.multiple_results = False


@register_primitive(
    primitive_obj=reshape_p,
    binding_factory=lambda: jnp.reshape,
    jaxpr_primitive=reshape_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.reshape.html",
    onnx=[
        {
            "component": "Reshape",
            "doc": "https://onnx.ai/onnx/operators/onnx__Reshape.html",
        }
    ],
    since="v0.1.0",
    context="primitives.jnp",
    component="reshape",
    testcases=[
        {
            "testcase": "reshape_1",
            "callable": lambda a: jnp.reshape(a, (2, 6)),
            "input_shapes": [(3, 4)],
        },
        {
            "testcase": "reshape_2",
            "callable": lambda a: jnp.reshape(a, (-1, 2)),
            "input_shapes": [(3, 4)],
        },
        {
            "testcase": "reshape_3",
            "callable": lambda a: jnp.reshape(a, (2, -1)),
            "input_shapes": [(3, 4)],
        },
        {
            "testcase": "reshape_4",
            "callable": lambda a: jnp.reshape(a, (-1, 4)),
            "input_shapes": [("B", 3, 4)],
        },
        {
            "testcase": "reshape_to_scalar",
            "callable": lambda a: jnp.reshape(a, ()),
            "input_shapes": [(1,)],
        },
        {
            "testcase": "reshape_from_scalar",
            "callable": lambda a: jnp.reshape(a, (1,)),
            "input_shapes": [()],
        },
        {
            "testcase": "reshape_cnn",
            "callable": lambda x: x.reshape(x.shape[0], -1),
            "input_shapes": [("B", 64, 14, 14)],
        },
    ],
)
class ReshapePlugin(PrimitiveLeafPlugin):
    """
    Plugin for converting jax.numpy.reshape to ONNX following the eval_shape pattern.
    """

    _ORIG_CALL: Callable[..., Any] | None = None

    # ------------------------------------------------------------
    # abstract_eval – delegate to original call via jax.eval_shape
    # ------------------------------------------------------------
    @staticmethod
    def abstract_eval(a: core.ShapedArray, *, newshape: Sequence[int | DimExpr], **_):
        """Use jax.eval_shape on the original jnp.reshape."""
        if ReshapePlugin._ORIG_CALL is None:
            raise RuntimeError("Original jnp.reshape not captured by ReshapePlugin.")

        def _helper(arr):
            # Need to handle the case where newshape might contain DimExpr directly
            # during abstract eval. JAX's original reshape should handle this.
            return ReshapePlugin._ORIG_CALL(arr, newshape)

        spec_a = jax.ShapeDtypeStruct(a.shape, a.dtype)
        out_spec = jax.eval_shape(_helper, spec_a)
        return core.ShapedArray(out_spec.shape, out_spec.dtype)

    # ------------------------------------------------------------
    # ONNX Conversion Logic (to_onnx)
    # ------------------------------------------------------------
    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of the reshape primitive to ONNX format."""
        input_var = node_inputs[0]
        output_var = node_outputs[0]

        if "new_sizes" in params:
            target_shape_param = params["new_sizes"]
        elif "newshape" in params:
            target_shape_param = params["newshape"]
        else:
            raise KeyError(
                "Could not find 'new_sizes' or 'newshape' in reshape parameters."
            )

        input_name = s.get_name(input_var)
        output_name = s.get_name(output_var)
        input_shape = input_var.aval.shape
        output_aval = output_var.aval
        output_shape = output_aval.shape  # Shape calculated by abstract_eval

        # --- Construct the target shape tensor RELIABLY ---
        shape_components = []
        # Ensure target_shape_param is treated as a sequence
        target_shape_seq = (
            target_shape_param
            if isinstance(target_shape_param, Sequence)
            else [target_shape_param]
        )
        # Handle empty tuple explicitly for reshape to scalar
        if (
            not isinstance(target_shape_param, Sequence)
            and target_shape_param is not None
        ):
            target_shape_seq = [target_shape_param]
        elif target_shape_param == ():  # Check for empty tuple specifically
            target_shape_seq = []
        else:
            # Cast to ensure type safety
            target_shape_seq = (
                list(target_shape_param) if target_shape_param is not None else []
            )

        symbolic_dim_map = getattr(
            s.builder, "reverse_map", getattr(s, "reverse_map", {})
        )
        symbol_to_index = {
            sym: idx for idx, sym in enumerate(input_shape) if isinstance(sym, DimExpr)
        }

        shape_of_input_name = None

        # --- Use target_shape_seq for iteration ---
        for dim in target_shape_seq:
            if isinstance(dim, DimExpr):
                axis_index = symbol_to_index.get(dim)
                if axis_index is None:
                    dim_str = symbolic_dim_map.get(dim, str(dim))
                    str_symbol_to_index = {
                        str(sym): idx for idx, sym in enumerate(input_shape)
                    }
                    axis_index = str_symbol_to_index.get(dim_str)
                    if axis_index is None:
                        raise ValueError(
                            f"Cannot map DimExpr {dim} ('{dim_str}') to input axis index in {input_shape}"
                        )

                if shape_of_input_name is None:
                    shape_of_input_name = s.get_unique_name(f"{input_name}_shape")
                    s.add_node(
                        helper.make_node(
                            "Shape",
                            inputs=[input_name],
                            outputs=[shape_of_input_name],
                            name=s.get_unique_name(f"shape_of_{input_name}"),
                        )
                    )
                    s.add_shape_info(shape_of_input_name, (len(input_shape),), np.int64)

                axis_const = s.get_constant_name(np.array(axis_index, dtype=np.int64))
                gathered_dim_scalar_int64 = s.get_unique_name(
                    f"{input_name}_dim{axis_index}_int64"
                )
                s.add_node(
                    helper.make_node(
                        "Gather",
                        inputs=[shape_of_input_name, axis_const],
                        outputs=[gathered_dim_scalar_int64],
                        axis=0,
                        name=s.get_unique_name(f"gather_dim_{axis_index}"),
                    )
                )
                s.add_shape_info(gathered_dim_scalar_int64, (), np.int64)

                # --- REMOVED INTERMEDIATE CASTS ---

                # --- CORRECTED UNSQUEEZE ---
                # Create constant tensor for 'axes' input
                unsqueeze_axes_const = s.get_constant_name(
                    np.array([0], dtype=np.int64)
                )
                unsqueezed_dim_name = s.get_unique_name(
                    f"{gathered_dim_scalar_int64}_unsqueezed"
                )
                # Provide 'axes' as the second input, remove 'axes' attribute
                s.add_node(
                    helper.make_node(
                        "Unsqueeze",
                        inputs=[
                            gathered_dim_scalar_int64,
                            unsqueeze_axes_const,
                        ],  # Data and Axes inputs
                        outputs=[unsqueezed_dim_name],
                        name=s.get_unique_name(
                            f"unsqueeze_{gathered_dim_scalar_int64}"
                        ),
                        # No 'axes' attribute here for opset >= 13
                    )
                )
                # --- END CORRECTED UNSQUEEZE ---
                s.add_shape_info(unsqueezed_dim_name, (1,), np.int64)
                shape_components.append(unsqueezed_dim_name)

            elif isinstance(dim, (int, np.integer)):
                const_name = s.get_constant_name(np.array([int(dim)], dtype=np.int64))
                shape_components.append(const_name)
            else:
                # This case should ideally not be hit if target_shape_seq is correct
                raise TypeError(
                    f"Unsupported dimension type: {type(dim)} in {target_shape_seq}"
                )

        # Concatenate components
        if not shape_components:  # Reshape to scalar case ()
            shape_tensor_name = s.get_constant_name(np.array([], dtype=np.int64))
        elif len(shape_components) == 1:
            shape_tensor_name = shape_components[0]
        else:
            shape_tensor_name = s.get_unique_name(f"{input_name}_target_shape")
            s.add_node(
                helper.make_node(
                    "Concat",
                    inputs=shape_components,
                    outputs=[shape_tensor_name],
                    axis=0,
                    name=s.get_unique_name("concat_target_shape"),
                )
            )
            s.add_shape_info(
                shape_tensor_name, (len(target_shape_seq),), np.int64
            )  # Use length of original sequence

        # --- Data Input/Output Type Handling (Keep previous logic) ---
        input_dtype = input_var.aval.dtype
        input_dtype_enum = s._ensure_onnx_dtype(input_dtype)
        expected_output_dtype = output_aval.dtype
        expected_output_dtype_enum = s._ensure_onnx_dtype(expected_output_dtype)
        reshape_data_input_name = input_name
        reshape_data_input_dtype_enum = input_dtype_enum
        if input_dtype_enum == TensorProto.INT32:
            casted_input_name = s.get_unique_name(f"{input_name}_casted_int64")
            s.add_node(
                helper.make_node(
                    "Cast",
                    inputs=[input_name],
                    outputs=[casted_input_name],
                    to=int(TensorProto.INT64),
                    name=s.get_unique_name("cast_reshape_data_to_int64"),
                )
            )
            s.add_shape_info(casted_input_name, input_shape, np.int64)
            reshape_data_input_name = casted_input_name
            reshape_data_input_dtype_enum = TensorProto.INT64

        # --- Emit ONNX Reshape node ---
        temp_reshape_output_name = s.get_unique_name(f"{input_name}_reshaped")
        reshape_node = helper.make_node(
            "Reshape",
            inputs=[reshape_data_input_name, shape_tensor_name],
            outputs=[temp_reshape_output_name],
            name=s.get_unique_name("reshape_0"),
            allowzero=0,
        )
        s.add_node(reshape_node)
        s.add_shape_info(
            temp_reshape_output_name, output_shape, reshape_data_input_dtype_enum
        )

        # --- Cast Output back to JAX expected type if necessary ---
        if reshape_data_input_dtype_enum != expected_output_dtype_enum:
            cast_node = helper.make_node(
                "Cast",
                inputs=[temp_reshape_output_name],
                outputs=[output_name],
                to=int(expected_output_dtype_enum),
                name=s.get_unique_name("cast_reshape_output_to_jax_expected"),
            )
            s.add_node(cast_node)
            s.add_shape_info(output_name, output_shape, expected_output_dtype_enum)
        else:
            identity_node = helper.make_node(
                "Identity",
                inputs=[temp_reshape_output_name],
                outputs=[output_name],
                name=s.get_unique_name("identity_reshape_output"),
            )
            s.add_node(identity_node)
            s.add_shape_info(output_name, output_shape, expected_output_dtype_enum)

    # ------------------------------------------------------------
    # monkey-patch – capture original & inject primitive binding
    # ------------------------------------------------------------
    @staticmethod
    def _reshape_binding(a, newshape, order="C"):
        """Binds inputs to the reshape primitive."""
        if order != "C":
            raise NotImplementedError("Only C-style reshape is supported.")
        return reshape_p.bind(a, newshape=newshape)

    @staticmethod
    def get_monkey_patch(orig_fn: Callable):
        """Returns the patched function that captures the original and binds the primitive."""
        ReshapePlugin._ORIG_CALL = orig_fn

        def patched_reshape(a, newshape, order="C"):
            return ReshapePlugin._reshape_binding(a, newshape, order)

        return patched_reshape

    @staticmethod
    def patch_info():
        """Provides patching information for jnp.reshape."""
        return {
            "patch_targets": [jnp],
            "target_attribute": "reshape",
            "patch_function": ReshapePlugin.get_monkey_patch,
        }


# --- Existing registration ---
reshape_p.def_abstract_eval(ReshapePlugin.abstract_eval)
# --- End Existing registration ---



from typing import TYPE_CHECKING

from flax import nnx
from jax.extend.core import Primitive
from onnx import helper

from jax2onnx.plugin_system import PrimitiveLeafPlugin, register_primitive

if TYPE_CHECKING:
    from jax2onnx.converter.jaxpr_converter import Jaxpr2OnnxConverter

# Define the GELU primitive
nnx.gelu_p = Primitive("nnx.gelu")
nnx.gelu_p.multiple_results = False  # Correct initialization


@register_primitive(
    jaxpr_primitive=nnx.gelu_p.name,
    jax_doc="https://jax.readthedocs.io/en/latest/_autosummary/jax.nn.gelu.html",
    onnx=[
        {
            "component": "Gelu",
            "doc": "https://onnx.ai/onnx/operators/onnx__Gelu.html",
        }
    ],
    since="v0.1.0",
    context="primitives.nnx",
    component="gelu",
    testcases=[
        {
            "testcase": "gelu",
            "callable": lambda x: nnx.gelu(x, approximate=False),
            "input_shapes": [(1,)],
        },
        {
            "testcase": "gelu_1",
            "callable": lambda x: nnx.gelu(x, approximate=False),
            "input_shapes": [(1, 10)],
        },
        {
            "testcase": "gelu_2",
            "callable": lambda x: nnx.gelu(x, approximate=True),
            "input_shapes": [(1,)],
        },
        {
            "testcase": "gelu_3",
            "callable": lambda x: nnx.gelu(x, approximate=True),
            "input_shapes": [("B", 10)],
        },
    ],
)
class GeluPlugin(PrimitiveLeafPlugin):
    """
    Plugin for converting flax.nnx.gelu to ONNX.
    """

    @staticmethod
    def abstract_eval(x, approximate=True):
        """Abstract evaluation function for GELU."""
        # Use update instead of creating a new ShapedArray to avoid issues with unhashable tracers
        return x.update(shape=x.shape, dtype=x.dtype, weak_type=False)

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handles conversion of GELU to ONNX format."""
        input_var = node_inputs[0]
        output_var = node_outputs[0]

        input_name = s.get_name(input_var)
        output_name = s.get_name(output_var)

        # Retrieve the approximate parameter (defaulting to True if not provided)
        approximate = params.get("approximate", True)
        approximation = "tanh" if approximate else "none"

        gelu_node = helper.make_node(
            "Gelu",
            inputs=[input_name],
            outputs=[output_name],
            name=s.get_unique_name("gelu"),
            approximate=approximation,  # Correctly pass 'approximation'
        )
        s.add_node(gelu_node)

    @staticmethod
    def get_monkey_patch():
        """Provides patching information for GELU."""

        def patched_gelu(x, approximate=True):
            return nnx.gelu_p.bind(x, approximate=approximate)

        return patched_gelu

    @staticmethod
    def patch_info():
        """Provides patching information for GELU."""
        return {
            "patch_targets": [nnx],
            "patch_function": lambda _: GeluPlugin.get_monkey_patch(),
            "target_attribute": "gelu",
        }


# Register abstract evaluation function
nnx.gelu_p.def_abstract_eval(GeluPlugin.abstract_eval)



from typing import TYPE_CHECKING

import jax
from onnx import helper

from jax2onnx.plugin_system import PrimitiveLeafPlugin, register_primitive

if TYPE_CHECKING:
    from jax2onnx.converter.jaxpr_converter import Jaxpr2OnnxConverter


@register_primitive(
    jaxpr_primitive=jax.lax.ne_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.lax.ne.html",
    onnx=[
        {
            "component": "Equal",
            "doc": "https://onnx.ai/onnx/operators/onnx__Equal.html",
        },
        {
            "component": "Not",
            "doc": "https://onnx.ai/onnx/operators/onnx__Not.html",
        },
    ],
    since="v0.2.0",
    context="primitives.lax",
    component="ne",
    testcases=[
        {
            "testcase": "ne",
            "callable": lambda x1, x2: x1 != x2,
            "input_shapes": [(3,), (3,)],
        }
    ],
)
class NePlugin(PrimitiveLeafPlugin):
    """Plugin for converting jax.lax.ne to ONNX Equal and Not."""

    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        """Handle JAX ne primitive."""
        input_names = [s.get_name(inp) for inp in node_inputs]
        eq_output = s.get_unique_name("equal_output")
        output_name = s.get_var_name(node_outputs[0])

        # Add value info for eq_output
        s.add_shape_info(eq_output, shape=node_inputs[0].aval.shape, dtype="bool")

        node_1 = helper.make_node(
            "Equal",
            inputs=input_names,
            outputs=[eq_output],
            name=s.get_unique_name("ne_eq"),
        )
        s.add_node(node_1)

        node_2 = helper.make_node(
            "Not",
            inputs=[eq_output],
            outputs=[output_name],
            name=s.get_unique_name("ne_not"),
        )
        s.add_node(node_2)



# file: jax2onnx/plugins/jax/numpy/concatenate.py

# --- Imports ---------------------------------------------------------------
from __future__ import annotations
from typing import TYPE_CHECKING, Callable, Any, Sequence

import jax
import jax.numpy as jnp
from jax import core
from jax.extend.core import Primitive
from onnx import helper

from jax2onnx.plugin_system import PrimitiveLeafPlugin, register_primitive
from jax2onnx.converter.patched_callable_wrapper import PatchedCallableWrapper

import logging

logger = logging.getLogger("jax2onnx.plugins.jax.numpy.concatenate")

_SENTINEL = -1  # what `jnp.tile` produces for “unknown” sizes


if TYPE_CHECKING:
    from jax2onnx.converter.jaxpr_converter import Jaxpr2OnnxConverter


def concat_dynamic_tile_func(x):
    # x : (B, N, D)
    D = x.shape[2]  # 256 (concrete) in the testcase
    token = jnp.zeros((1, 1, D), dtype=x.dtype)  # (1, 1, D)

    # broadcast_to accepts symbolic sizes, so we can use `x.shape[0]`
    tiled_token = jnp.broadcast_to(token, (x.shape[0], 1, D))  # (B, 1, D)

    return jnp.concatenate([tiled_token, x], axis=1)  # (B, 1+N, D)


# ---------------------------------------------------------------------------
#  Primitive definition
# ---------------------------------------------------------------------------
if not hasattr(jnp, "concatenate_p"):
    jnp.concatenate_p = Primitive("jnp.concatenate")
    jnp.concatenate_p.multiple_results = False


# ---------------------------------------------------------------------------
#  Plugin
# ---------------------------------------------------------------------------
@register_primitive(
    jaxpr_primitive=jnp.concatenate_p.name,
    jax_doc="https://docs.jax.dev/en/latest/_autosummary/jax.numpy.concatenate.html",
    onnx=[
        {
            "component": "Concat",
            "doc": "https://onnx.ai/onnx/operators/onnx__Concat.html",
        }
    ],
    since="v0.1.0",
    context="primitives.jnp",
    component="concatenate",
    testcases=[
        {
            "testcase": "concatenate",
            "callable": lambda a, b: jnp.concatenate((a, b), axis=0),
            "input_shapes": [(3,), (3,)],
        },
        {
            "testcase": "concatenate_abstract_middle_dim",
            "callable": lambda a, b: jnp.concatenate((a, b), axis=1),
            "input_shapes": [("B", 1, 8), ("B", 10, 8)],
            "expected_output_shapes": [("B", 11, 8)],
        },
        {
            "testcase": "concatenate_tile_and_symbolic",
            "callable": concat_dynamic_tile_func,
            "input_shapes": [("B", 49, 256)],  # Matches failing ConcatClsToken
            "expected_output_shapes": [("B", 50, 256)],  # 1 + 49 = 50
        },
    ],
)
class ConcatenatePlugin(PrimitiveLeafPlugin):
    """
    Symbolic-shape aware converter for `jax.numpy.concatenate`.
    Its `abstract_eval` rule defers shape inference to **jax.eval_shape**,
    which is safe to call while the outer `jax.make_jaxpr` trace is live.
    """

    # Will be filled the first time we patch `jnp.concatenate`
    _ORIGINAL_CONCATENATE: Callable | None = None

    # ---------------------------------------------------------------------
    #  abstract_eval  (⇐ **now uses jax.eval_shape**)
    # ---------------------------------------------------------------------
    @staticmethod
    def abstract_eval(*avals: core.ShapedArray, axis: int):
        logger.debug("ConcatenatePlugin.abstract_eval – start")

        # ---- sanity checks ------------------------------------------------
        if not avals:
            raise ValueError("concatenate expects at least one input")
        if not all(isinstance(a, core.ShapedArray) for a in avals):
            raise TypeError(
                "All inputs to concatenate must be ShapedArray, got "
                f"{[type(a) for a in avals]}"
            )
        if not isinstance(axis, int):
            raise TypeError(f"`axis` must be an int, got {type(axis)}")

        # ---- original function reference ---------------------------------
        orig = ConcatenatePlugin._ORIGINAL_CONCATENATE
        if orig is None:
            raise RuntimeError("Original jnp.concatenate was not captured.")

        # ---- ShapeDtypeStruct specs --------------------------------------
        specs = [jax.ShapeDtypeStruct(a.shape, a.dtype) for a in avals]

        # ---- helper that calls the *un-patched* concatenate --------------
        def _helper(*xs):
            return orig(xs, axis=axis)

        # ---- delegate to jax.eval_shape ----------------------------------
        try:
            result_spec = jax.eval_shape(_helper, *specs)
            result_spec = jax.tree_util.tree_leaves(result_spec)[0]
            return core.ShapedArray(result_spec.shape, result_spec.dtype)
        except Exception as exc:
            logger.debug("eval_shape failed, using manual rule: %s", exc)
            shape = ConcatenatePlugin._manual_shape(avals, axis=axis)
            dtype = jax.dtypes.result_type(*[a.dtype for a in avals])
            return core.ShapedArray(shape, dtype)

    # ---------------------------------------------------------------------
    #  to_onnx – unchanged
    # ---------------------------------------------------------------------
    def to_onnx(self, s: "Jaxpr2OnnxConverter", node_inputs, node_outputs, params):
        axis = int(params.get("axis", 0))
        node = helper.make_node(
            "Concat",
            inputs=[s.get_name(v) for v in node_inputs],
            outputs=[s.get_name(node_outputs[0])],
            name=s.get_unique_name("concat"),
            axis=axis,
        )
        s.add_node(node)

        out_aval = node_outputs[0].aval
        s.add_shape_info(
            s.get_name(node_outputs[0]), tuple(out_aval.shape), out_aval.dtype
        )

    # ---------------------------------------------------------------------
    #  patch_info – capture original fn & inject wrapper
    # ---------------------------------------------------------------------
    @staticmethod
    def patch_info() -> dict[str, Any]:
        def _creator(orig_fn: Callable):
            logger.info("Storing original jnp.concatenate reference")
            ConcatenatePlugin._ORIGINAL_CONCATENATE = orig_fn
            return PatchedCallableWrapper(orig_fn, jnp.concatenate_p)

        return {
            "patch_targets": [jnp],
            "patch_function": _creator,
            "target_attribute": "concatenate",
        }

    @staticmethod
    def _manual_shape(avals: Sequence[core.ShapedArray], *, axis: int):
        """Light‑weight concatenate shape rule that tolerates the -1 sentinel."""
        rank = len(avals[0].shape)
        out: list[Any] = []  # Add type annotation to match expected list type

        for d in range(rank):
            if d == axis:
                # sum along the concat axis, ignoring sentinels
                sizes = [a.shape[d] for a in avals]
                int_total = sum(
                    s for s in sizes if isinstance(s, int) and s != _SENTINEL
                )
                sym_sizes = [
                    s for s in sizes if not isinstance(s, int) or s == _SENTINEL
                ]
                if sym_sizes:
                    # any symbolic → keep the symbolic one (they must all agree)
                    out.append(sym_sizes[0])
                else:
                    out.append(int_total)
            else:
                # all other axes must agree up to broadcasting of 1 / sentinel
                size_set = {
                    s for s in (a.shape[d] for a in avals) if s not in (1, _SENTINEL)
                }
                if len(size_set) > 1:
                    raise TypeError("non‑concat dims disagree: " + str(size_set))
                out.append(next(iter(size_set)) if size_set else avals[0].shape[d])
        return tuple(out)


# Register the rule with the primitive
jnp.concatenate_p.def_abstract_eval(ConcatenatePlugin.abstract_eval)



"""
JAXPR to ONNX Converter Module

This module contains the core functionality for converting JAX's JAXPR representation
to ONNX format. It provides the main Jaxpr2OnnxConverter class which traverses the JAXPR
representation of a JAX function and converts it to equivalent ONNX operations.
"""

from typing import Any, Dict, List, Union
import logging
import jax
import jax.random
import numpy as np
from jax.extend import core as extend_core
from jax.extend.core import Literal, ClosedJaxpr
from onnx import helper
from jax2onnx.converter.onnx_builder import OnnxBuilder
from jax2onnx.converter.monkey_patch_utils import temporary_monkey_patches
from jax2onnx.plugin_system import (
    ONNX_FUNCTION_PLUGIN_REGISTRY,
    PLUGIN_REGISTRY,
    PrimitiveLeafPlugin,
    import_all_plugins,
)
from jax._src.export.shape_poly import _DimExpr
from jax import ShapeDtypeStruct

logger = logging.getLogger("jax2onnx.converter.jaxpr_converter")


class Jaxpr2OnnxConverter:
    """
    Converts JAX's JAXPR representation to ONNX format, enabling interoperability
    between JAX and ONNX-based tools.

    This class handles the core conversion logic from JAX's internal representation
    to the ONNX graph format. It traverses the JAXPR computation graph and
    generates equivalent ONNX operations.
    """

    # Map symbolic dimensions to their origin tensor names and axes

    # mapping from ONNX tensor name → its symbolic shape (a tuple of ints or dim-names)
    symbolic_shapes: dict[str, tuple[Union[int, str], ...]]

    def __init__(self, builder: OnnxBuilder):
        self.logger = logging.getLogger("jax2onnx.converter.jaxpr_converter")
        self.builder = builder
        setattr(self.builder, "converter", self)
        self.params: Dict[str, Any] = {}
        self.call_params: Dict[str, Any] = {}
        self.var_to_name: dict[Any, str] = {}
        self.name_to_var: dict[str, Any] = {}
        self.primitive_handlers: dict[str, Any] = {}
        self.shape_env: dict[str, tuple[int, ...]] = {}
        self.name_to_const: dict[str, Any] = {}
        # ------------------------------------------------------------------
        # Mapping: symbolic dimension expression -> (tensor_name, axis)
        # ------------------------------------------------------------------
        self.symbolic_dim_to_origin: dict[_DimExpr, tuple[str, int]] = {}
        import_all_plugins()
        self._register_primitive_handlers()
        self.symbolic_shapes = {}
        # Initialize DimVar -> Name mappings needed for dim_to_symbol
        self._dimvar_to_name = getattr(builder, "var_to_symbol_map", {})
        self._dimvar_to_name_by_str = {
            str(k): v for k, v in self._dimvar_to_name.items()
        }
        # Note: dim_to_symbol might need access to the builder's map directly later
        # self.dim_to_symbol = lambda d: _canonical_symbol(self.builder, d) # Maybe pass builder?

    def new_var(self, dtype: np.dtype, shape: tuple[int, ...]) -> extend_core.Var:
        """Create a new JAX variable with the given dtype and shape."""
        return extend_core.Var(
            self.builder.get_unique_name(""), extend_core.ShapedArray(shape, dtype)
        )

    def add_node(self, node: Any) -> None:
        """Add an ONNX node to the builder."""
        self.builder.add_node(node)

    def _emit_result(self, jax_outvar, wanted_sym, src_sym):
        """
        Connect a JAX output variable to its ONNX representation.

        If the caller asked for a specific output symbol name that's different
        from the source symbol, insert an Identity node to create the alias.

        Args:
            jax_outvar: The JAX output variable
            wanted_sym: The desired output symbol name
            src_sym: The source symbol name currently producing the value
        """
        # If the caller asked for a different symbol, materialise an Identity
        if wanted_sym != src_sym:
            id_name = self.builder.name_generator.get("alias")
            self.builder.add_node(
                helper.make_node("Identity", [src_sym], [wanted_sym], name=id_name)
            )
            self.builder.add_value_info(
                wanted_sym, jax_outvar.aval.shape, jax_outvar.aval.dtype
            )
            self.var_to_name[jax_outvar] = wanted_sym
        else:
            self.var_to_name[jax_outvar] = src_sym

    def _import_var_as_input(
        self,
        var: extend_core.Var,
        sym: str,
    ):
        """Make *sym* a formal graph input **iff** it is not
             already created inside the builder**.

        That is the case when *sym* is produced by a node that
        was added to ``self.builder`` **before** this converter
        started processing the JAXPR – for example the ``Cast``
        that makes ``iter64 → iter32`` in a Loop body.  Turning
        such an internal tensor into another graph input would
        (a) pollute the interface and (b) create duplicate-name
        errors at ONNX runtime.
        """

        # 1. Fast bail-out if another node already writes *sym*
        if any(sym in n.output for n in self.builder.nodes):
            # We only need shape-information (if not yet there)
            aval = var.aval
            self.builder.add_value_info(sym, aval.shape, aval.dtype)
            return

        # 2. Really add a new formal input
        aval = var.aval
        self.builder.add_input(sym, aval.shape, aval.dtype)

    def get_unique_name(self, prefix: str = "node") -> str:
        """Get a unique name for an ONNX node or variable."""
        return self.builder.get_unique_name(prefix)

    def get_var_name(self, var: Any) -> str:
        """Get or create a unique name for a JAX variable."""

        # ────────────────────────────────────────────────────────────────
        # Plain Python int / float (e.g., literal `3`)
        # ────────────────────────────────────────────────────────────────
        from numbers import Number

        if isinstance(var, Number) and not hasattr(var, "aval"):
            value = np.array(var, dtype=np.int32)
            const_name = self.get_constant_name(value)
            wanted_dtype = self._ensure_onnx_dtype(value.dtype)

            # Cast to ensure dtype matches int32 exactly
            cast_name = self.get_unique_name("lit_cast_to_i32")
            self.builder.add_node(
                helper.make_node(
                    "Cast",
                    inputs=[const_name],
                    outputs=[cast_name],
                    to=int(wanted_dtype),
                    name=cast_name,
                )
            )
            self.add_shape_info(cast_name, (), value.dtype)
            return cast_name

        # ────────────────────────────────────────────────────────────────
        # Handle Literal (JAX constant-folded literals)
        # ────────────────────────────────────────────────────────────────
        if isinstance(var, Literal):
            value = np.asarray(var.val)
            if np.issubdtype(value.dtype, np.integer):
                value = value.astype(np.int32)
            const_name = self.get_constant_name(value)
            self.add_shape_info(const_name, value.shape, value.dtype)
            return const_name

        # ────────────────────────────────────────────────────────────────
        # Every other normal Var
        # ────────────────────────────────────────────────────────────────
        if var not in self.var_to_name:
            name = self.get_unique_name("var")
            self.set_var_name(var, name)
        return self.var_to_name[var]

    def set_var_name(self, var: Any, name: str) -> None:
        """Set a custom name for a JAX variable."""
        self.var_to_name[var] = name
        self.name_to_var[name] = var

    def change_var_name(self, var: Any, name: str) -> None:
        """Change the name of a JAX variable."""
        if var in self.var_to_name:
            old_name = self.var_to_name[var]
            del self.name_to_var[old_name]
            self.var_to_name[var] = name
            self.name_to_var[name] = var
            self.builder.change_var_name(old_name, name)
        else:
            raise ValueError(f"Variable {var} not found in var_to_name mapping.")

    def get_constant_name(self, val: Any) -> str:
        """Get or create a name for a constant value in the ONNX graph."""
        return self.builder.get_constant_name(val)

    def _ensure_onnx_dtype(self, dtype):
        """
        Ensure the dtype is a valid ONNX TensorProto data type (integer).

        Args:
            dtype: The data type to convert (numpy.dtype, Python type, or ONNX enum)

        Returns:
            An integer representing an ONNX TensorProto data type
        """
        from onnx import TensorProto

        # Centralized mapping for numpy and string dtypes
        dtype_map = {
            np.float32: TensorProto.FLOAT,
            np.float64: TensorProto.DOUBLE,
            np.int32: TensorProto.INT32,
            np.int64: TensorProto.INT64,
            np.bool_: TensorProto.BOOL,
            np.uint8: TensorProto.UINT8,
            np.int8: TensorProto.INT8,
            np.uint16: TensorProto.UINT16,
            np.int16: TensorProto.INT16,
            np.uint32: TensorProto.UINT32,
            np.uint64: TensorProto.UINT64,
            np.float16: TensorProto.FLOAT16,
            np.complex64: TensorProto.COMPLEX64,
            np.complex128: TensorProto.COMPLEX128,
            "float32": TensorProto.FLOAT,
            "float64": TensorProto.DOUBLE,
            "int32": TensorProto.INT32,
            "int64": TensorProto.INT64,
            "bool": TensorProto.BOOL,
            "uint8": TensorProto.UINT8,
            "int8": TensorProto.INT8,
            "uint16": TensorProto.UINT16,
            "int16": TensorProto.INT16,
            "uint32": TensorProto.UINT32,
            "uint64": TensorProto.UINT64,
            "float16": TensorProto.FLOAT16,
            "complex64": TensorProto.COMPLEX64,
            "complex128": TensorProto.COMPLEX128,
        }

        # If it's already an int, assume it's a valid ONNX enum
        if isinstance(dtype, int):
            return dtype

        # Handle JAX array types
        if hasattr(dtype, "__module__") and dtype.__module__.startswith("jax"):
            if "int" in str(dtype):
                return TensorProto.INT64
            elif "float" in str(dtype):
                return TensorProto.FLOAT
            elif "bool" in str(dtype):
                return TensorProto.BOOL

        # Handle numpy dtypes and string names
        if hasattr(dtype, "type") and dtype.type in dtype_map:
            return dtype_map[dtype.type]
        if hasattr(dtype, "name") and dtype.name in dtype_map:
            return dtype_map[dtype.name]
        if isinstance(dtype, str) and dtype in dtype_map:
            return dtype_map[dtype]

        # Try ONNX's helper (might raise TypeError for some inputs)
        try:
            return helper.np_dtype_to_tensor_dtype(dtype)
        except (TypeError, ValueError):
            self.logger.debug(
                "Could not convert dtype %s to ONNX dtype, defaulting to FLOAT", dtype
            )
            return TensorProto.FLOAT

    def _handle_pjit(self, eqn, params):
        """Inline a `pjit` call inside the current graph."""

        # ① fetch the closed jaxpr (API changed a few times)
        closed = params.get("call_jaxpr") or params.get("jaxpr")
        if isinstance(closed, ClosedJaxpr):
            inner_jaxpr = closed.jaxpr
            consts = closed.consts
        else:  # already an open jaxpr
            inner_jaxpr = closed
            consts = params.get("consts", ())

        # ② recursively convert the body
        self._process_jaxpr(inner_jaxpr, consts)

        # ③ wire inner outputs → outer outputs
        for outer_var, inner_var in zip(eqn.outvars, inner_jaxpr.outvars):
            inner_name = self.get_name(inner_var)
            self.set_var_name(outer_var, inner_name)

    def register_shape(self, name: str, shape: tuple[int, ...], dtype: Any) -> str:
        """Register shape and dtype information for a tensor, preserving symbolic dims."""
        # Convert dtype to ONNX TensorProto enum if needed
        onnx_dtype = self._ensure_onnx_dtype(dtype)

        # If the shape comes from a ShapeDtypeStruct or similar, preserve symbolic tokens
        # Try to recover symbolic names if present (e.g., from .symbol attribute or original spec)
        symbolic_shape = tuple(d.symbol if hasattr(d, "symbol") else d for d in shape)

        # Register with the builder
        self.builder.register_value_info_metadata(name, symbolic_shape, onnx_dtype)

        # Store locally for quick access
        self.shape_env[name] = symbolic_shape

        return name

    def add_input(self, var: Any, shape: tuple, dtype: Any = np.float32) -> str:
        name = self.get_var_name(var)
        self.builder.add_input(
            name, shape, dtype
        )  # Pass potentially symbolic shape tuple
        return name

    def add_output(self, var: Any, shape: tuple, dtype: Any = np.float32) -> str:
        name = self.get_var_name(var)
        self.builder.add_output(
            name, shape, dtype
        )  # Pass potentially symbolic shape tuple
        return name

    def get_name(self, var: Any) -> str:
        """Get the ONNX name for a JAX variable or literal."""
        if isinstance(var, jax._src.core.Var):
            return self.get_var_name(var)
        elif isinstance(var, extend_core.Literal):
            return self.get_constant_name(var)
        else:
            raise NotImplementedError("not yet implemented")

    def _extract_symbolic_axes(self, example_args):
        # Returns a tuple of all symbolic dimension tokens in the example args (if any)
        symbolic_axes = set()
        for arg in example_args:
            if hasattr(arg, "shape"):
                for d in arg.shape:
                    if not isinstance(d, int):
                        symbolic_axes.add(d)
        # JAX expects a tuple, not a set, for abstracted_axes
        return tuple(symbolic_axes) if symbolic_axes else None

    def dim_to_symbol(self, d):
        """Translate JAX shape-dimension `d` into a stable symbolic string
        or a concrete int."""
        # 0) plain integer → just return it
        if isinstance(d, int):
            return d

        # 1) exact identity hit
        if d in getattr(self, "_dimvar_to_name", {}):
            return self._dimvar_to_name[d]

        # 2) hit by (count, dtype) — survives Var cloning
        key = (getattr(d, "count", None), str(getattr(d, "aval", "")))
        if key in getattr(self, "_dimvar_to_name_by_count", {}):
            return self._dimvar_to_name_by_count[key]

        # 3) modern JAX: DimExpr carries .symbol
        if hasattr(d, "symbol") and d.symbol is not None:
            return str(d.symbol)

        # 4) fall back to old helper

        _logger = logging.getLogger("jax2onnx.converter.jaxpr_converter")
        _logger.debug("  - FALLBACK to _symbol_name: %s ⚠️", d)

        # try to reuse an existing symbol (same position in the arg-shape)
        sym = None
        if (
            hasattr(self.builder, "current_arg_axes")
            and self.builder.current_arg_axes is not None
        ):
            # current_arg_axes e.g. (None, 'B', None)
            idx = getattr(self.builder, "current_axis_index", 0)  # maintained by caller
            if idx < len(self.builder.current_arg_axes):
                sym = self.builder.current_arg_axes[idx]  # 'B' or None

        if not sym:  # still nothing? invent one
            if hasattr(self.builder, "_unique_symbol"):
                sym = self.builder._unique_symbol()  # e.g. '__sym0'
            else:
                sym = f"__sym{id(d) % 1000}"  # fallback if _unique_symbol doesn't exist

        # register every alias so the object can be found again later
        if not hasattr(self.builder, "var_to_symbol_map"):
            self.builder.var_to_symbol_map = {}

        self.builder.var_to_symbol_map[d] = sym
        self.builder.var_to_symbol_map[id(d)] = sym
        self.builder.var_to_symbol_map[str(d)] = sym

        logger.debug("[dim_to_symbol] %s (%s)  →  %s", d, type(d).__name__, sym)
        return sym  # <— now make_value_info gets "B" (or '__sym0')

        # Step	Description	Dynamic Dim Handling
        # -	User provides symbolic dimensions ("B")	User-level symbolic dimension
        # -	Map symbolic dimensions to concrete ints	Temporary numeric placeholders
        # -	Create concrete zero-arrays for JAX tracer	Concrete numeric array
        # -	Trace with abstracted_axes	JAX records symbolic shapes (DimVar)
        # -	Extract symbolic shapes post-tracing	Explicit symbolic shapes recorded
        # -	Export symbolic shapes into ONNX	ONNX dynamic shape (dim_param)

    ###############################################################################
    # NOTE: this *replaces* the old trace_jaxpr implementation
    ###############################################################################
    # --- Helper for safe dimension-to-symbol conversion (Keep from response #33) ---
    def _dim_to_symbol_safe(self, d):
        if isinstance(d, int):
            return d
        # Use the builder's map which should be up-to-date
        # Try object, then str representation
        resolved = self.builder.var_to_symbol_map.get(d)
        if resolved is None:
            resolved = self.builder.var_to_symbol_map.get(str(d))
        if resolved is None:
            # Fallback for unknown symbolic objects (might be internal JAX exprs)
            logger.warning(
                f"Cannot resolve symbolic dim {d} (type: {type(d)}) to name. Using str()."
            )
            resolved = str(d)  # Use string representation as fallback name
        return resolved

    # --- MODIFIED trace_jaxpr Method ---
    def trace_jaxpr(
        self,
        fn: Any,
        # Change signature: Now expects the list of symbolic ShapeDtypeStruct avals
        symbolic_avals: List[ShapeDtypeStruct],  # Changed name and type hint
        preserve_graph: bool = False,
        params: dict[str, Any] | None = None,
    ) -> None:
        """
        Trace a JAX function to JAXPR using pre-computed symbolic abstract values
        and convert it to ONNX.
        """
        self.logger.debug(
            f"trace_jaxpr called with {len(symbolic_avals)} symbolic avals. preserve_graph={preserve_graph}"
        )
        if not preserve_graph:
            # Reset state for top-level call
            # NOTE: Ensure builder.reset() doesn't clear var_to_symbol_map or re-assign it after reset
            # Maybe builder reset needs adjustment, or we fetch the map *after* reset?
            # Let's assume builder keeps the map or we re-set it from conversion_api if needed.
            # self.builder.reset() # Defer reset or handle map persistence carefully
            self.var_to_name.clear()
            self.name_to_const.clear()
            self.shape_env.clear()  # This seems safe to clear
            # Fetch map from builder, assuming it was set in conversion_api
            self._dimvar_to_name = getattr(self.builder, "var_to_symbol_map", {})
            self._dimvar_to_name_by_str = {
                str(k): v for k, v in self._dimvar_to_name.items()
            }
            self.symbolic_shapes.clear()
        else:
            # For nested calls, inherit maps - ensure builder map is current context
            self._dimvar_to_name = getattr(self.builder, "var_to_symbol_map", {})
            self._dimvar_to_name_by_str = {
                str(k): v for k, v in self._dimvar_to_name.items()
            }
            # Keep existing self.symbolic_shapes for nested scope? This needs thought.

        # --- Step 1: Use received symbolic_avals directly ---
        # No need to create concrete tracing_args based on symbolic_dim_map
        # Always hand ShapeDtypeStructs to jax.make_jaxpr
        tracing_args = [
            (
                jax.ShapeDtypeStruct(a.shape, a.dtype)  # ✔ keeps _DimExpr symbols
                if hasattr(a, "shape")
                and hasattr(a, "dtype")  # Check for ShapedArray attributes
                else a
            )  # already a ShapeDtypeStruct
            for a in symbolic_avals
        ]
        self.logger.debug(f"Using tracing_args (symbolic avals): {tracing_args}")

        # --- Step 2: Call jax.make_jaxpr ---
        # We *should not* need abstracted_axes if symbolic shapes are explicit in avals
        # JAX handles polymorphism based on the symbolic objects in the input avals.
        with temporary_monkey_patches(allow_function_primitives=True):
            # Remove abstracted_axes argument if it was previously used
            mk = jax.make_jaxpr(fn)
            try:
                # Pass symbolic avals directly as arguments
                closed = mk(*tracing_args, **(params or {}))
            except Exception as e:
                self.logger.error(
                    f"jax.make_jaxpr failed with symbolic avals. Error: {e}",
                    exc_info=True,
                )
                self.logger.error(f"Function: {fn}")
                self.logger.error(f"Tracing Args (Symbolic Avals): {tracing_args}")
                self.logger.error(f"Params: {params}")
                raise

        self.logger.debug(f"Jaxpr generated: {closed}")
        self.jaxpr = closed.jaxpr
        self.output_vars = getattr(self.jaxpr, "outvars", [])  # Access outvars safely

        # --------------------------------------------------------------
        # Record symbolic dimensions present in the graph inputs
        # --------------------------------------------------------------
        for input_var, input_spec in zip(self.jaxpr.invars, symbolic_avals):
            tensor_name = self.get_name(input_var)
            for axis, dim in enumerate(input_spec.shape):
                if isinstance(dim, _DimExpr):
                    self.symbolic_dim_to_origin[dim] = (tensor_name, axis)

        # --- Step 3: Post-trace Processing (Update internal state) ---
        # Ensure the builder has the necessary map for subsequent operations
        # It should have been set in conversion_api.py
        self.builder.var_to_symbol_map = self._dimvar_to_name

        # Store symbolic shapes for *all* vars seen in the final jaxpr
        # using the safe conversion back to string names
        self.symbolic_shapes = {}
        all_vars = (
            getattr(self.jaxpr, "invars", [])
            + getattr(self.jaxpr, "constvars", [])
            + getattr(self.jaxpr, "outvars", [])
        )
        for var in all_vars:
            if var is not None and hasattr(var, "aval") and hasattr(var.aval, "shape"):
                try:
                    name = self.get_name(var)  # Handles Vars and Literals
                    # Convert potentially symbolic shape back to tuple with string names
                    sym_shape = tuple(
                        self._dim_to_symbol_safe(d) for d in var.aval.shape
                    )
                    self.symbolic_shapes[name] = sym_shape
                    self.logger.debug(
                        f"Stored symbolic shape for {name} ('{type(var)}'): {sym_shape}"
                    )
                except Exception as e:
                    # Log details about the variable causing issues
                    var_repr = repr(var)
                    aval_repr = repr(getattr(var, "aval", None))
                    self.logger.warning(
                        f"Could not process/store symbolic shape for var {var_repr} with aval {aval_repr}. Error: {e}",
                        exc_info=False,
                    )  # Keep log concise

        # --- Step 4: Convert JAXPR to ONNX Graph ---
        self.logger.info("Processing generated jaxpr...")
        # Pass the map explicitly or ensure _process_jaxpr uses self.builder.var_to_symbol_map
        self._process_jaxpr(self.jaxpr, closed.consts)
        self.logger.info("Jaxpr processing complete.")

    def _process_jaxpr(self, jaxpr: Any, consts: list[Any]) -> None:
        # --------------------------------------------------------------------
        # 1) Special‐case: static‐only JAXPR (e.g. `lambda x: x.shape[0]` ⇒ literal)
        # --------------------------------------------------------------------
        if not jaxpr.eqns and len(jaxpr.invars) == 1 and len(jaxpr.outvars) == 1:
            jaxpr.invars[0]
            out = jaxpr.outvars[0]
            # Only proceed if the JAXPR actually returns a literal int
            from jax.extend.core import Literal

            if isinstance(out, Literal) or isinstance(out, (int,)):
                # pull out the integer
                static_val = int(out.val if isinstance(out, Literal) else out)
                # --- emit exactly the static‐dim ONNX snippet ---
                import numpy as _np
                from onnx import TensorProto

                # 1) make int64 initializer
                arr = _np.array(static_val, dtype=_np.int64)
                init_name = self.get_constant_name(arr)

                # 2) Cast it down to INT32
                cast_name = self.get_unique_name("dim_as_value_static_cast")
                self.builder.add_node(
                    helper.make_node(
                        "Cast",
                        inputs=[init_name],
                        outputs=[cast_name],
                        to=int(TensorProto.INT32),
                        name=cast_name,
                    )
                )
                # record its shape (scalar) and dtype
                self.add_shape_info(cast_name, (), _np.int32)

                # 3) Identity into the real output var
                out_name = self.get_var_name(out)
                id_name = self.get_unique_name("dim_as_value_static_id")
                self.builder.add_node(
                    helper.make_node(
                        "Identity", inputs=[cast_name], outputs=[out_name], name=id_name
                    )
                )
                self.add_shape_info(out_name, (), _np.int32)
                # 4) register it as graph output
                self.builder.add_output(out_name, (), _np.int32)
                return

        # --------------------------------------------------------------------
        # 2) Otherwise, normal path: register any true JAXPR‐constvars...
        # --------------------------------------------------------------------
        for i, const in enumerate(consts):
            # register initializer-name → value
            const_name = self.get_constant_name(const)
            const_var = jaxpr.constvars[i]
            self.var_to_name[const_var] = const_name
            self.name_to_var[const_name] = const_var
            self.name_to_const[const_name] = const
            # also register the "logical var name" (in case it differs)
            var_name = self.get_var_name(const_var)
            if var_name != const_name:
                self.name_to_const[var_name] = const

        # Add input variables with symbolic shapes
        for var in jaxpr.invars:
            if var is None:
                continue
            var_name = self.get_var_name(var)
            if not hasattr(var, "aval") or not hasattr(var.aval, "shape"):
                continue
            shape = tuple(self._dim_to_symbol_safe(d) for d in var.aval.shape)
            dtype = var.aval.dtype
            if not any(inp.name == var_name for inp in self.builder.inputs):
                self.add_input(var, shape, dtype)

        # Process equations
        for eqn in jaxpr.eqns:
            self._process_eqn(eqn)

        # Explicitly handle outputs
        for var in jaxpr.outvars:
            if var is None:
                continue
            name = self.get_var_name(var)

            if name in self.name_to_const:
                # Explicitly wrap initializer in Identity for valid ONNX output
                identity_output_name = self.get_unique_name("const_output")
                dtype = np.asarray(self.name_to_const[name]).dtype
                shape = ()

                identity_node = helper.make_node(
                    "Identity",
                    inputs=[name],
                    outputs=[identity_output_name],
                    name=self.get_unique_name("identity_const_out"),
                )
                self.builder.add_node(identity_node)

                # Set the output explicitly to the Identity node's output
                self.builder.add_output(identity_output_name, shape, dtype)
                self.add_shape_info(identity_output_name, shape, dtype)
            elif hasattr(var, "aval") and hasattr(var.aval, "shape"):
                # if the plugin already emitted this output, skip it
                if any(o.name == name for o in self.builder.outputs):
                    continue
                shape = tuple(self._dim_to_symbol_safe(d) for d in var.aval.shape)
                dtype = var.aval.dtype
                self.add_output(var, shape, dtype)
            else:
                if not any(o.name == name for o in self.builder.outputs):
                    self.builder.add_output(name, (), np.int32)

    def add_shape_info(self, name: str, shape: tuple, dtype: Any = np.float32) -> str:
        # Note: shape passed here might already be symbolic strings from _dim_to_symbol_safe
        self.builder.add_value_info(name, shape, dtype)
        return name

    def _create_identity_node(
        self, node_inputs: list[Any], node_outputs: list[Any], prefix: str
    ) -> Any:
        """Create an Identity node to handle simple pass-through operations."""

        input_name = self.get_name(node_inputs[0])
        output_name = self.get_var_name(node_outputs[0])

        node = helper.make_node(
            "Identity",
            inputs=[input_name],
            outputs=[output_name],
            name=self.get_unique_name(f"{prefix}:identity"),
        )
        self.builder.add_node(node)
        return node

    def _register_primitive_handlers(self) -> None:
        """Register all primitive handlers from both plugin registries."""
        # Register handlers from the main plugin registry
        for key, plugin in PLUGIN_REGISTRY.items():
            if isinstance(plugin, PrimitiveLeafPlugin):
                self.primitive_handlers[key] = plugin.get_handler(self)

        # Register handlers from the ONNX function plugin registry
        for plugin in ONNX_FUNCTION_PLUGIN_REGISTRY.values():
            primitive = plugin.primitive
            self.primitive_handlers[primitive.name] = plugin.get_handler(self)

        # built‑in call‑style primitive that we inline
        self.primitive_handlers["pjit"] = lambda conv, eqn, params: conv._handle_pjit(
            eqn, params
        )

        if self.primitive_handlers:
            self.logger.debug(
                "Registered %d primitive handlers", len(self.primitive_handlers)
            )

    def _process_eqn(self, eqn: Any) -> None:
        """Process a single JAXPR equation by dispatching to the appropriate plugin handler."""

        if not hasattr(eqn, "primitive"):
            # Should not happen for standard jaxprs, maybe handle call primitives?
            self.logger.warning(
                f"Equation type without 'primitive' attribute encountered: {type(eqn)}. Skipping: {eqn}"
            )
            return

        primitive = eqn.primitive
        name = primitive.name

        # Check if it's handled by function plugins first (if applicable)
        is_function_handler = (
            name in ONNX_FUNCTION_PLUGIN_REGISTRY
        )  # Use the actual registry name

        handler = self.primitive_handlers.get(name)
        if handler is None:
            raise NotImplementedError(
                f"No ONNX handler registered for JAX primitive: '{name}'"
            )

        self.logger.debug(f"Processing eqn for primitive: {name}")
        try:
            # Call the handler (typically a method on the plugin instance or a lambda)
            # The handler expects: self (converter), eqn object, params dict
            handler(self, eqn, eqn.params)  # Pass self, eqn, params
        except Exception as e:
            self.logger.error(
                f"Error processing primitive '{name}' with handler {handler}. Eqn: {eqn}",
                exc_info=True,
            )
            raise RuntimeError(f"Failed processing primitive '{name}'") from e

        # --- Handle output shapes for non-function primitives ---
        # (This logic might need refinement based on how function plugin outputs are handled)
        # Register shape info for the output variables of this equation
        # Use the safe dim-to-symbol conversion
        if not is_function_handler:  # Only for leaf primitives?
            for outvar in eqn.outvars:
                if (
                    outvar is not None
                    and hasattr(outvar, "aval")
                    and hasattr(outvar.aval, "shape")
                ):
                    output_name = self.get_name(outvar)  # Get ONNX name
                    # Use the potentially symbolic shape from the aval
                    shape_tuple = tuple(
                        self._dim_to_symbol_safe(d) for d in outvar.aval.shape
                    )
                    dtype = outvar.aval.dtype
                    # Add shape info to the builder
                    self.add_shape_info(output_name, shape_tuple, dtype)
                else:
                    # Handle literals or vars without shape info if necessary
                    pass



# In file: jax2onnx/converter/conversion_api.py

# Add Tuple, Union to imports if not already present
from typing import Any, Dict, Sequence, Union, List, Tuple  # Added Tuple, List
import onnx
import logging
import numpy as np
from onnx import helper, mapping
from jax2onnx.converter.dynamic_utils import (
    _create_symbolic_input_avals,
)  # Import the helper
from jax2onnx.converter.jaxpr_converter import Jaxpr2OnnxConverter
from jax2onnx.converter.name_generator import UniqueNameGenerator
from jax2onnx.converter.onnx_builder import OnnxBuilder
from jax2onnx.converter.optimize_onnx_graph import improve_onnx_model
import jax.numpy as jnp

logger = logging.getLogger("jax2onnx.converter.conversion_api")


# Remove or comment out the old prepare_example_args if no longer needed
# def prepare_example_args(...): ...


# ------------------------------------------------------------------
# Promote items passed via *input_params* to proper graph inputs
# ------------------------------------------------------------------
def _elem_type_from_numpy(arr: np.ndarray) -> int:
    return mapping.NP_TYPE_TO_TENSOR_TYPE[arr.dtype]


def _promote_params_to_inputs(model: onnx.ModelProto, params: Dict[str, Any] | None):
    if not params:
        return

    for name, value in params.items():
        # ① drop initializer (if any)
        kept = [init for init in model.graph.initializer if init.name != name]
        model.graph.ClearField("initializer")
        model.graph.initializer.extend(kept)

        # ② drop stale value_info (INT32 in our case)
        kept = [vi for vi in model.graph.value_info if vi.name != name]
        model.graph.ClearField("value_info")
        model.graph.value_info.extend(kept)

        # ③ add graph input once
        if any(inp.name == name for inp in model.graph.input):
            continue
        dtype = _elem_type_from_numpy(np.asarray(value))
        vi = helper.make_tensor_value_info(name, dtype, [])  # scalar
        model.graph.input.append(vi)


# -----------------------------------------------------------------------------
# drop duplicate initialisers for parameters promoted to real graph inputs
# -----------------------------------------------------------------------------
def _strip_param_initializers(model, input_params):
    if not input_params:
        return
    param_names = set(input_params)
    keep = [init for init in model.graph.initializer if init.name not in param_names]
    del model.graph.initializer[:]  # in‑place update
    model.graph.initializer.extend(keep)


def to_onnx(
    fn: Any,
    # Assume 'inputs' is passed as a list/sequence of shape tuples
    inputs: Sequence[Sequence[Union[int, str]]],
    input_params: Dict[str, Any] | None = None,
    model_name: str = "jax_model",
    opset: int = 21,
    default_dtype: Any = jnp.float32,  # Default dtype if not specified otherwise
    # ... other parameters ...
) -> onnx.ModelProto:
    """
    Converts a JAX function into an ONNX model.
    Handles symbolic dimensions specified as strings in input shapes.
    """
    logger.info(f"Starting JAX to ONNX conversion for '{model_name}'")
    logger.debug(f"Received raw inputs (shapes): {inputs}")
    logger.debug(
        f"Received input_params: {input_params.keys() if input_params else 'None'}"
    )
    logger.debug(f"Using default dtype: {default_dtype}")

    # --- Step 0: Format input_specs ---
    # Create the list of (shape, dtype) tuples needed by the helper
    # This assumes 'inputs' is a list of shapes and uses default_dtype.
    # Future enhancement: Allow user to pass [(shape, dtype), ...] directly.
    try:
        input_specs: List[Tuple[Sequence[Union[int, str]], Any]] = (
            []
        )  # Use List and Tuple
        for shape_spec in inputs:

            shape_tuple: Tuple[Union[int, str], ...] = tuple(shape_spec)
            # Pair the processed shape tuple with the default dtype
            input_specs.append((shape_tuple, default_dtype))
    except Exception as e:
        logger.error(
            f"Failed to format input shapes/dtypes. Input: {inputs}. Error: {e}",
            exc_info=True,
        )
        raise TypeError(
            "Input shapes must be sequences (tuples/lists) of int or str."
        ) from e

    logger.debug(f"Formatted input_specs: {input_specs}")

    # --- Step 1: Prepare Abstract Inputs with Symbolic Dimensions ---
    # (Assumes this part is now correct)
    symbolic_avals, var_to_symbol_map = _create_symbolic_input_avals(input_specs)

    # --- Setup Converter and Builder ---
    unique_name_generator = UniqueNameGenerator()

    # Initialize OnnxBuilder *without* the map argument
    builder = OnnxBuilder(
        unique_name_generator,
        opset=opset,
        converter=None,
        # Removed var_to_symbol_name_map from here
    )
    # Set the map as an attribute *after* initialization
    builder.var_to_symbol_map = var_to_symbol_map
    logger.debug(f"Set builder.var_to_symbol_map: {builder.var_to_symbol_map}")

    # Initialize Converter and link back (no change here)
    converter = Jaxpr2OnnxConverter(builder)
    builder.converter = converter

    converter.call_params = input_params or {}

    # --- Step 2: Trace the function using Symbolic Avals ---
    # Reminder: converter.trace_jaxpr needs modification next
    logger.info("Initiating JAX tracing with symbolic abstract values...")
    # *** NEXT STEP: Modify converter.trace_jaxpr to accept symbolic_avals ***
    converter.trace_jaxpr(fn, symbolic_avals, params=input_params)
    logger.info("JAX tracing finished.")

    # --- Step 3: Build and Optimize ONNX model ---
    logger.info("Building ONNX model...")
    builder.filter_unused_initializers()
    model = builder.create_onnx_model(model_name)

    # Replace with the new function for properly handling parameter inputs
    _promote_params_to_inputs(
        model, input_params
    )  # ← new instead of _strip_param_initializers

    logger.info("Optimizing ONNX model...")
    model = improve_onnx_model(model)
    logger.info("ONNX model conversion complete.")

    logger.debug(onnx.helper.printable_graph(model.graph))

    return model


def analyze_constants(model: onnx.ModelProto):
    """
    Analyzes constants in an ONNX model and prints a detailed report.

    This function is useful for debugging and understanding how constants are
    represented and used within the ONNX graph.

    Args:
        model: The ONNX model to analyze.
    """
    logger.info("\n🔍 Constant Analysis Report (Verbose)")
    graph = model.graph
    graph_inputs = {inp.name for inp in graph.input}
    initializers = {init.name for init in graph.initializer}
    const_nodes = {
        node.output[0]: node for node in graph.node if node.op_type == "Constant"
    }
    function_names = {f.name for f in model.functions}
    logger.info("\n📦 Top-Level Inputs:")
    for inp in graph.input:
        logger.info(f"  - {inp.name}")
    logger.info("\n🧊 Initializers (Style 2):")
    for init in graph.initializer:
        logger.info(f"  - {init.name}")
    logger.info("\n🧱 Constant Nodes in Main Graph (Style 2):")
    for name in const_nodes:
        logger.info(f"  - {name}")
    logger.info("\n🧩 Function Call Inputs:")
    for node in graph.node:
        if node.op_type in function_names:
            logger.info(f"\n▶ Function Call: {node.op_type}")
            for inp in node.input:
                style = "Unknown/Intermediate"
                if inp in initializers:
                    style = "Style 2 (initializer reused)"
                elif inp in graph_inputs:
                    style = "Style 1 (passed in as input)"
                elif inp in const_nodes:
                    style = "Style 2 (constant node)"
                logger.info(f"  - {inp} → {style}")
    logger.info("\n🔗 Constant Usage Map:")
    for node in graph.node:
        for inp in node.input:
            if inp.startswith("const_") or inp.startswith("var_"):
                logger.info(f"  - {inp} used in {node.op_type}")



"""
ONNX Function Builder Module

This module provides specialized functionality for building ONNX functions.
It works with OnnxBuilder but focuses specifically on the function-related aspects
of ONNX model construction.
"""

from typing import Any, Dict, List, Optional

from onnx import FunctionProto, NodeProto, ValueInfoProto, helper

from jax2onnx.converter.name_generator import UniqueNameGenerator
from jax2onnx.converter.onnx_builder import CUSTOM_DOMAIN


class OnnxFunctionBuilder:
    """
    A specialized builder for ONNX functions, handling the creation,
    registration, and management of ONNX function definitions.

    This class separates function-specific building logic from the general
    ONNX model building functionality to improve code organization.
    """

    def __init__(self, onnx_builder: Optional[Any] = None):
        """
        Initialize an ONNX function builder.

        Args:
            onnx_builder: Optional OnnxBuilder instance to integrate with.
                          If provided, this function builder will use the same
                          name generator and can access the OnnxBuilder's context.
        """
        self.functions: Dict[str, FunctionProto] = {}
        self.onnx_builder = onnx_builder

        # Use shared name generator if onnx_builder is provided, otherwise create new one
        if onnx_builder is not None:
            self.name_generator = onnx_builder.name_generator
        else:
            self.name_generator = UniqueNameGenerator()

    def register_function(self, function_proto: FunctionProto) -> None:
        """
        Register an ONNX function with the builder.

        Args:
            function_proto: The FunctionProto object to register.
        """
        self.functions[function_proto.name] = function_proto

    def create_function(
        self,
        domain: str,
        name: str,
        inputs: List[str],
        outputs: List[str],
        nodes: List[NodeProto],
        opset: int,
        value_info: Optional[List[ValueInfoProto]] = None,
    ) -> FunctionProto:
        """
        Create an ONNX function with the specified parameters.

        Args:
            domain: The domain of the function.
            name: The name of the function.
            inputs: The input names of the function.
            outputs: The output names of the function.
            nodes: The nodes contained in the function.
            opset: The opset version to use.
            value_info: Optional value info for intermediate tensors.

        Returns:
            An ONNX FunctionProto object.
        """
        function_proto = helper.make_function(
            domain=domain,
            fname=name,
            inputs=inputs,
            outputs=outputs,
            nodes=nodes,
            opset_imports=[
                helper.make_opsetid("", opset),
                helper.make_opsetid(domain, 1),
            ],
            value_info=value_info or [],
        )
        self.register_function(function_proto)
        return function_proto

    def get_unique_function_name(self, base_name: str) -> str:
        """
        Get a unique function name based on the provided base name.

        Args:
            base_name: The base name for the function.

        Returns:
            A unique function name.
        """
        return self.name_generator.get(base_name)

    def get_all_functions(self) -> List[FunctionProto]:
        """
        Get all registered functions, ensuring no duplicates.

        Returns:
            A list of unique FunctionProto objects.
        """
        # Use a dictionary to filter duplicates by name
        unique_functions = {f.name: f for f in self.functions.values()}
        return list(unique_functions.values())

    def create_function_call_node(
        self,
        function_name: str,
        input_names: List[str],
        output_names: List[str],
        domain: str = CUSTOM_DOMAIN,
        node_name: Optional[str] = None,
        op_type: Optional[str] = None,
    ) -> NodeProto:
        """
        Create a node that calls an ONNX function.

        Args:
            function_name: The name of the function to call.
            input_names: The input tensor names.
            output_names: The output tensor names.
            domain: The function domain.
            node_name: Optional name for the node.
            op_type: Optional operation type for the node.

        Returns:
            A NodeProto representing a function call.
        """
        if node_name is None:
            node_name = self.name_generator.get(function_name.split(".")[-1])

        return helper.make_node(
            op_type=op_type or function_name,
            inputs=input_names,
            outputs=output_names,
            name=node_name,
            domain=domain,
        )

    def merge_functions(self, other_functions: Dict[str, FunctionProto]) -> None:
        """
        Merge functions from another source into this builder.

        Args:
            other_functions: Dictionary of functions to merge.
        """
        import warnings

        for name, func in other_functions.items():
            if name not in self.functions:
                self.functions[name] = func
            else:
                warnings.warn(
                    f"[Duplicate function] Skipping already-registered function '{name}'",
                    RuntimeWarning,
                )



# file: jax2onnx/plugins/examples/nnx/fori_loop.py

import jax
from jax2onnx.plugin_system import register_example


def model_fn(x):
    steps = 5

    def body_func(index, args):
        x, counter = args
        x += 0.1 * x**2
        counter += 1
        return (x, counter)

    args = (x, 0)
    y, _ = jax.lax.fori_loop(0, steps, body_func, args)

    return y


register_example(
    component="ForiLoop",
    description="fori_loop example",
    since="v0.5.1",
    context="examples.nnx",
    children=[],
    testcases=[
        {
            "testcase": "fori_loop_counter",
            "callable": lambda x: model_fn(x),
            "input_shapes": [(1,)],
        },
    ],
)


# main function
# def main():
#     x = jnp.array([1.0, 2.0], dtype=jnp.float32)
#     y = model_fn(x)


# if __name__ == "__main__":
#     main()
